@@ -18,7 +18,7 @@
 /*****************************************************************************/
 
 #include "gftp.h"
-static const char cvsid[] = "$Id: config_file.c,v 1.31 2003/07/23 02:39:23 masneyb Exp $";
+static const char cvsid[] = "$Id: config_file.c,v 1.38 2004/03/17 13:03:24 masneyb Exp $";
 
 void
 gftp_add_bookmark (gftp_bookmarks_var * newentry)
@@ -105,6 +105,60 @@ gftp_add_bookmark (gftp_bookmarks_var * newentry)
 }
 
 
+static int
+copyfile (char *source, char *dest)
+{
+  int srcfd, destfd;
+  char buf[8192];
+  ssize_t n;
+
+  if ((srcfd = gftp_fd_open (NULL, source, O_RDONLY, 0)) == -1)
+    {
+      printf (_("Error: Cannot open local file %s: %s\n"),
+              source, g_strerror (errno));
+      exit (1);
+    }
+
+  if ((destfd = gftp_fd_open (NULL, dest, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR)) == -1)
+    {
+      printf (_("Error: Cannot open local file %s: %s\n"),
+              dest, g_strerror (errno));
+      close (srcfd);
+      exit (1);
+    }
+
+  while ((n = read (srcfd, buf, sizeof (buf))) > 0)
+    {
+      if (write (destfd, buf, n) == -1)
+        {
+          printf (_("Error: Could not write to socket: %s\n"), 
+                  g_strerror (errno));
+          exit (1);
+        }
+    }
+
+  if (n == -1)
+    {
+      printf (_("Error: Could not read from socket: %s\n"), g_strerror (errno));
+      exit (1);
+    }
+
+  if (close (srcfd) == -1)
+    {
+      printf (_("Error closing file descriptor: %s\n"), g_strerror (errno));
+      exit (1);
+    }
+
+  if (close (destfd) == -1)
+    {
+      printf (_("Error closing file descriptor: %s\n"), g_strerror (errno));
+      exit (1);
+    }
+
+  return (1);
+}
+
+
 static void
 gftp_read_bookmarks (char *global_data_path)
 {
@@ -204,7 +258,10 @@ gftp_read_bookmarks (char *global_data_path)
 	  curpos = buf + 9;
 	  if (newentry->pass)
 	    g_free (newentry->pass);
-	  newentry->pass = g_strdup (curpos);
+
+	  /* Always try to descramble passords. If the password is not
+             scrambled, descramble_password returns the string unchanged */
+	  newentry->pass = gftp_descramble_password (curpos);
 	  newentry->save_password = *newentry->pass != '\0';
 	}
       else if (strncmp (buf, "account", 7) == 0 && newentry)
@@ -660,12 +717,13 @@ write_comment (FILE * fd, const char *comment)
 void
 gftp_write_bookmarks_file (void)
 {
+  char *bmhdr, *pwhdr, *tempstr, *password, buf[256];
   gftp_bookmarks_var * tempentry;
-  char *bmhdr, *tempstr;
   FILE * bmfile;
   int i;
 
   bmhdr = N_("Bookmarks file for gFTP. Copyright (C) 1998-2003 Brian Masney <masneyb@gftp.org>. Warning: Any comments that you add to this file WILL be overwritten");
+  pwhdr = N_("Note: The passwords contained inside this file are scrambled. This algorithm is not secure. This is to avoid your password being easily remembered by someone standing over your shoulder while you're editing this file. Prior to this, all passwords were stored in plaintext.");
 
   if ((tempstr = expand_path (BOOKMARKS_FILE)) == NULL)
     {
@@ -683,6 +741,7 @@ gftp_write_bookmarks_file (void)
   g_free (tempstr);
 
   write_comment (bmfile, _(bmhdr));
+  write_comment (bmfile, _(pwhdr));
   fwrite ("\n", 1, 1, bmfile);
 
   tempentry = gftp_bookmarks->children;
@@ -693,9 +752,16 @@ gftp_write_bookmarks_file (void)
 	  tempentry = tempentry->children;
 	  continue;
 	}
+
       tempstr = tempentry->path;
       while (*tempstr == '/')
 	tempstr++;
+
+      if (tempentry->save_password && tempentry->pass != NULL)
+        password = gftp_scramble_password (tempentry->pass);
+      else
+        password = NULL;
+
       fprintf (bmfile,
 	       "[%s]\nhostname=%s\nport=%d\nprotocol=%s\nremote directory=%s\nlocal directory=%s\nusername=%s\npassword=%s\naccount=%s\n",
 	       tempstr, tempentry->hostname == NULL ? "" : tempentry->hostname,
@@ -705,17 +771,20 @@ gftp_write_bookmarks_file (void)
 	       tempentry->remote_dir == NULL ? "" : tempentry->remote_dir,
 	       tempentry->local_dir == NULL ? "" : tempentry->local_dir,
 	       tempentry->user == NULL ? "" : tempentry->user,
-	       !tempentry->save_password
-	       || tempentry->pass == NULL ? "" : tempentry->pass,
+	       password == NULL ? "" : password,
 	       tempentry->acct == NULL ? "" : tempentry->acct);
 
+      if (password != NULL)
+        g_free(password);
+
       if (tempentry->local_options_vars != NULL)
         {
           for (i=0; i<tempentry->num_local_options_vars; i++)
             {
-              fprintf (bmfile, "%s=", tempentry->local_options_vars[i].key);
-              gftp_option_types[tempentry->local_options_vars[i].otype].write_function (&tempentry->local_options_vars[i], bmfile, 1);
-              fprintf (bmfile, "\n");
+              gftp_option_types[tempentry->local_options_vars[i].otype].write_function (&tempentry->local_options_vars[i], buf, sizeof (buf), 1);
+
+              fprintf (bmfile, "%s=%s\n", tempentry->local_options_vars[i].key,
+                       buf);
             }
         }
 
@@ -739,10 +808,10 @@ gftp_write_bookmarks_file (void)
 void
 gftp_write_config_file (void)
 {
+  char *tempstr, buf[256];
   gftp_config_vars * cv;
   GList *templist;
   FILE *conffile;
-  char *tempstr;
   int i;
 
   if ((tempstr = expand_path (CONFIG_FILE)) == NULL)
@@ -778,9 +847,9 @@ gftp_write_config_file (void)
           if (cv[i].comment != NULL)
             write_comment (conffile, _(cv[i].comment));
 
-          fprintf (conffile, "%s=", cv[i].key);
-          gftp_option_types[cv[i].otype].write_function (&cv[i], conffile, 1);
-          fprintf (conffile, "\n");
+          gftp_option_types[cv[i].otype].write_function (&cv[i], buf,
+                                                         sizeof (buf), 1);
+          fprintf (conffile, "%s=%s\n", cv[i].key, buf);
         }
     }
     
@@ -881,15 +950,15 @@ gftp_config_file_read_text (char *str, gftp_config_vars * cv, int line)
 
 
 static int
-gftp_config_file_write_text (gftp_config_vars * cv, FILE * fd, int to_config_file)
+gftp_config_file_write_text (gftp_config_vars * cv, char *buf, size_t buflen,
+                             int to_config_file)
 {
   char *outstr;
 
   if (cv->value != NULL)
     {
       outstr = cv->value;
-      if (*outstr != '\0')
-        fprintf (fd, "%s", outstr);
+      g_snprintf (buf, buflen, "%s", outstr);
       return (0);
     }
   else
@@ -898,7 +967,8 @@ gftp_config_file_write_text (gftp_config_vars * cv, FILE * fd, int to_config_fil
 
 
 static int
-gftp_config_file_write_hidetext (gftp_config_vars * cv, FILE * fd, int to_config_file)
+gftp_config_file_write_hidetext (gftp_config_vars * cv, char *buf,
+                                 size_t buflen, int to_config_file)
 {
   char *outstr;
 
@@ -908,9 +978,9 @@ gftp_config_file_write_hidetext (gftp_config_vars * cv, FILE * fd, int to_config
       if (*outstr != '\0')
         {
           if (to_config_file)
-            fprintf (fd, "%s", outstr);
+            g_snprintf (buf, buflen, "%s", outstr);
           else
-            fprintf (fd, "*****");
+            g_snprintf (buf, buflen, "*****");
         }
       return (0);
     }
@@ -969,9 +1039,10 @@ gftp_config_file_read_int (char *str, gftp_config_vars * cv, int line)
 
 
 static int
-gftp_config_file_write_int (gftp_config_vars * cv, FILE * fd, int to_config_file)
+gftp_config_file_write_int (gftp_config_vars * cv, char *buf, size_t buflen,
+                            int to_config_file)
 {
-  fprintf (fd, "%d", GPOINTER_TO_INT(cv->value));
+  g_snprintf (buf, buflen, "%d", GPOINTER_TO_INT (cv->value));
   return (0);
 }
 
@@ -994,21 +1065,22 @@ gftp_config_file_read_checkbox (char *str, gftp_config_vars * cv, int line)
 static int
 gftp_config_file_read_float (char *str, gftp_config_vars * cv, int line)
 {
-  float f;
+  union { intptr_t i; float f; } r;
 
-  f = strtod (str, NULL);
-  memcpy (&cv->value, &f, sizeof (cv->value));
+  r.f = strtod (str, NULL);
+  memcpy (&cv->value, &r.i, sizeof (cv->value));
   return (0);
 }
 
 
 static int
-gftp_config_file_write_float (gftp_config_vars * cv, FILE * fd, int to_config_file)
+gftp_config_file_write_float (gftp_config_vars * cv, char *buf, size_t buflen,
+                              int to_config_file)
 {
   float f;
 
   memcpy (&f, &cv->value, sizeof (f));
-  fprintf (fd, "%.2f", f);
+  g_snprintf (buf, buflen, "%.2f", f);
   return (0);
 }
 
@@ -1051,12 +1123,13 @@ gftp_config_file_read_color (char *str, gftp_config_vars * cv, int line)
 
 
 static int
-gftp_config_file_write_color (gftp_config_vars * cv, FILE * fd, int to_config_file)
+gftp_config_file_write_color (gftp_config_vars * cv, char *buf, size_t buflen,
+                              int to_config_file)
 {
   gftp_color * color;
 
   color = cv->value;
-  fprintf (fd, "%x:%x:%x", color->red, color->green, color->blue);
+  g_snprintf (buf, buflen, "%x:%x:%x", color->red, color->green, color->blue);
   return (0);
 }
 
@@ -1111,15 +1184,16 @@ gftp_config_file_read_intcombo (char *str, gftp_config_vars * cv, int line)
 
 
 static int
-gftp_config_file_write_intcombo (gftp_config_vars * cv, FILE * fd, int to_config_file)
+gftp_config_file_write_intcombo (gftp_config_vars * cv, char *buf,
+                                 size_t buflen, int to_config_file)
 {
   char **clist;
 
   clist = cv->listdata;
   if (clist != NULL)
-    fprintf (fd, "%s", clist[GPOINTER_TO_INT(cv->value)]);
+    g_snprintf (buf, buflen, "%s", clist[GPOINTER_TO_INT(cv->value)]);
   else
-    fprintf (fd, _("<unknown>"));
+    g_snprintf (buf, buflen, _("<unknown>"));
 
   return (0);
 }
@@ -1181,7 +1255,7 @@ gftp_option_type_var gftp_option_types[] = {
 
 
 void
-gftp_lookup_global_option (char * key, void *value)
+gftp_lookup_global_option (const char * key, void *value)
 {
   gftp_config_list_vars * tmplistvar;
   gftp_config_vars * tmpconfigvar;
@@ -1202,7 +1276,8 @@ gftp_lookup_global_option (char * key, void *value)
 
 
 void
-gftp_lookup_request_option (gftp_request * request, char * key, void *value)
+gftp_lookup_request_option (gftp_request * request, const char * key,
+                            void *value)
 {
   gftp_config_vars * tmpconfigvar;
 
@@ -1216,7 +1291,8 @@ gftp_lookup_request_option (gftp_request * request, char * key, void *value)
 
 
 void
-gftp_lookup_bookmark_option (gftp_bookmarks_var * bm, char * key, void *value)
+gftp_lookup_bookmark_option (gftp_bookmarks_var * bm, const char * key,
+                             void *value)
 {
   gftp_config_vars * tmpconfigvar;
 
@@ -1230,7 +1306,7 @@ gftp_lookup_bookmark_option (gftp_bookmarks_var * bm, char * key, void *value)
 
 
 void
-gftp_set_global_option (char * key, const void *value)
+gftp_set_global_option (const char * key, const void *value)
 {
   gftp_config_vars * tmpconfigvar, newconfigvar;
   void *nc_ptr;
@@ -1280,7 +1356,8 @@ _gftp_set_option_value (gftp_config_vars * cv, const void * newval)
 
 
 void
-gftp_set_request_option (gftp_request * request, char * key, const void *value)
+gftp_set_request_option (gftp_request * request, const char * key,
+                         const void *value)
 {
   gftp_config_vars * tmpconfigvar;
 
@@ -1314,7 +1391,8 @@ gftp_set_request_option (gftp_request * request, char * key, const void *value)
 
 
 void
-gftp_set_bookmark_option (gftp_bookmarks_var * bm, char * key, const void *value)
+gftp_set_bookmark_option (gftp_bookmarks_var * bm, const char * key,
+                          const void *value)
 {
   gftp_config_vars * tmpconfigvar, newconfigvar;
   int ret;
@@ -1370,11 +1448,13 @@ gftp_register_config_vars (gftp_config_vars * config_vars)
 void
 gftp_copy_local_options (gftp_config_vars ** new_options_vars, 
                          GHashTable ** new_options_hash,
+                         int *new_num_local_options_vars,
                          gftp_config_vars * orig_options,
                          int num_local_options_vars)
 {
   int i;
 
+  *new_num_local_options_vars = num_local_options_vars;
   if (orig_options == NULL || num_local_options_vars == 0)
     {
       *new_options_vars = NULL;
