@@ -17,7 +17,7 @@
 /*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                */
 /*****************************************************************************/
 
-/* $Id: gftp.h,v 1.58 2003/08/21 20:24:08 masneyb Exp $ */
+/* $Id: gftp.h,v 1.88 2004/03/26 21:03:49 masneyb Exp $ */
 
 #ifndef __GFTP_H
 #define __GFTP_H
@@ -26,6 +26,15 @@
 #include "config.h"
 #endif
 
+/* *BSD include 64bit off_t but _LARGEFILE_SOURCE is not defined. */
+#if defined (SIZEOF_OFF_T) && SIZEOF_OFF_T > 4 && !defined (_LARGEFILE_SOURCE)
+#define _LARGEFILE_SOURCE 1
+#endif
+
+#if defined (_LARGEFILE_SOURCE) && !defined (_LARGEFILE64_SOURCE)
+#define _LARGEFILE64_SOURCE 1
+#endif
+
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
@@ -57,6 +66,10 @@
 #include <grp.h>
 #include <math.h>
 
+#ifdef HAVE_STDINT_H
+#include <stdint.h>
+#endif
+
 #ifdef HAVE_STRING_H
 #include <string.h>
 #else
@@ -91,6 +104,11 @@
 #define GFTP_GET_AI_FAMILY(request)	AF_INET
 #endif
 
+/* Solaris needs this included for major()/minor() */
+#ifdef HAVE_SYS_MKDEV_H
+#include <sys/mkdev.h>
+#endif
+
 /* We need the major() and minor() macros in the user interface. If they aren't
    defined by the system, we'll just define them here. */
 #ifndef major
@@ -129,6 +147,12 @@
 
 #endif
 
+#ifdef __GNUC__
+# define GFTP_LOG_FUNCTION_ATTRIBUTES __attribute__((format(printf, 3, 4)))
+#else
+# define GFTP_LOG_FUNCTION_ATTRIBUTES
+#endif
+
 /* Server types (used by FTP protocol from SYST command) */
 #define GFTP_DIRTYPE_UNIX	1
 #define GFTP_DIRTYPE_EPLF	2
@@ -137,10 +161,19 @@
 #define GFTP_DIRTYPE_DOS	5	
 #define GFTP_DIRTYPE_VMS	6	
 #define GFTP_DIRTYPE_OTHER 	7
+#define GFTP_DIRTYPE_MVS	8
 
 /* Error types */
-#define GFTP_ERETRYABLE		-1
-#define GFTP_EFATAL		-2
+#define GFTP_ERETRYABLE		-1		/* Temporary failure. The GUI
+						   should wait briefly */
+#define GFTP_EFATAL		-2		/* Fatal error */
+#define GFTP_ERETRYABLE_NO_WAIT	-3		/* Temporary failure. The GUI
+						   should not wait and should
+						   reconnect */
+#define GFTP_ENOTRANS		-4		/* Custom error. This is
+						   returned when a FXP transfer
+						   is requested */
+#define GFTP_ETIMEDOUT		-5		/* Connected timed out */
 
 /* Some general settings */
 #define BASE_CONF_DIR		"~/.gftp"
@@ -148,20 +181,15 @@
 #define BOOKMARKS_FILE		BASE_CONF_DIR "/bookmarks"
 #define LOG_FILE		BASE_CONF_DIR "/gftp.log"
 #define MAX_HIST_LEN		10
-
-typedef enum gftp_transfer_type_tag 
-{
-  gftp_transfer_passive,
-  gftp_transfer_active
-} gftp_transfer_type;
-
+#define GFTP_URL_USAGE		"[[protocol://][user[:pass]@]site[:port][/directory]]"
 
 typedef enum gftp_logging_level_tag 
 {
   gftp_logging_send,
   gftp_logging_recv,
   gftp_logging_error,
-  gftp_logging_misc
+  gftp_logging_misc,
+  gftp_logging_misc_nolog 	/* Log to the screen, but don't log to disk */
 } gftp_logging_level;
 
 typedef struct gftp_file_tag gftp_file;
@@ -170,9 +198,6 @@ typedef struct gftp_file_tag gftp_file;
 #define GFTP_TRANS_ACTION_RESUME		2
 #define GFTP_TRANS_ACTION_SKIP			3
 
-#define GFTP_DIRECTION_DOWNLOAD			0
-#define GFTP_DIRECTION_UPLOAD			1
-
 #define GFTP_SORT_COL_FILE			1
 #define GFTP_SORT_COL_SIZE			2
 #define GFTP_SORT_COL_USER			3
@@ -216,7 +241,7 @@ struct gftp_file_tag
 
 typedef struct gftp_proxy_hosts_tag 
 {
-  /* FIXME - add IPV4 stuff here */
+  /* FIXME - add IPV6 stuff here */
 
   gint32 ipv4_network_address, 
          ipv4_netmask;
@@ -281,7 +306,8 @@ typedef struct gftp_config_vars_tag
 typedef struct gftp_option_type_tag
 {
   int (*read_function) (char *str, gftp_config_vars * cv, int line);
-  int (*write_function) (gftp_config_vars * cv, FILE * fd, int to_config_file);
+  int (*write_function) (gftp_config_vars * cv, char *buf, size_t buflen,
+                         int to_config_file);
   void (*copy_function) (gftp_config_vars * cv, gftp_config_vars * dest_cv);
   int (*compare_function) (gftp_config_vars * cv1, gftp_config_vars * cv2);
   void *(*ui_print_function) (gftp_config_vars * cv, void *user_data, void *value);
@@ -345,8 +371,6 @@ struct gftp_request_tag
                need_hostport : 1,
                need_userpass : 1,
                use_cache : 1,           /* Enable or disable the cache */
-               use_threads : 1,         /* Whether we need to spawn a thread
-                                           for this protocol */
                cached : 1,              /* Is this directory listing cached? */
                cancel : 1,		/* If a signal is received, should
 					   we cancel this operation */
@@ -360,6 +384,8 @@ struct gftp_request_tag
   void *user_data;
 
   int (*init)				( gftp_request * request );
+  void (*copy_param_options) 		( gftp_request * dest_request, 
+					  gftp_request * src_request );
   ssize_t (*read_function)		( gftp_request * request,
 					  void *ptr, 
 					  size_t size, 
@@ -445,13 +471,13 @@ typedef struct gftp_transfer_tag
   gftp_request * fromreq,
                * toreq;
 
-  unsigned int transfer_direction : 1, /* FIXME - take out */
-               cancel : 1,
+  unsigned int cancel : 1,
                ready : 1,
                started : 1,
                done : 1,
                show : 1,
                stalled : 1,
+               conn_error_no_timeout : 1,
                next_file : 1,
                skip_file : 1;
 
@@ -470,6 +496,8 @@ typedef struct gftp_transfer_tag
        current_file_retries;
 
   off_t curtrans,		/* Current transfered bytes for this file */
+        tot_file_trans,		/* Total number of bytes in the file being
+                                   transferred */
         curresumed,		/* Resumed bytes for this file */
         trans_bytes,		/* Amount of data transfered for entire 
 				   transfer */
@@ -501,9 +529,12 @@ typedef struct supported_gftp_protocols_tag
   int (*init) (gftp_request * request);		/* Init function */
   void (*register_options) (void);		/* Protocol options */
   char *url_prefix;				/* URL Prefix */
-  int shown;					/* Whether this protocol is 
+  unsigned int shown : 1,			/* Whether this protocol is 
                                                    shown or not to the user in 
                                                    the protocol dropdown box */
+               use_threads : 1;			/* Whether or not operations in
+						   this protocol should use
+						   threads */
 } supported_gftp_protocols;
 
 
@@ -614,32 +645,33 @@ GHashTable * build_bookmarks_hash_table	( gftp_bookmarks_var * entry );
 
 void print_bookmarks 			( gftp_bookmarks_var * bookmarks );
 
-void gftp_lookup_global_option 		( char * key, 
+void gftp_lookup_global_option 		( const char * key, 
 					  void *value );
 
 void gftp_lookup_request_option 	( gftp_request * request, 
-					  char * key, 
+					  const char * key, 
 					  void *value );
 
 void gftp_lookup_bookmark_option 	( gftp_bookmarks_var * bm, 
-					  char * key, 
+					  const char * key, 
 					  void *value );
 
-void gftp_set_global_option 		( char * key, 
+void gftp_set_global_option 		( const char * key, 
 					  const void *value );
 
 void gftp_set_request_option 		( gftp_request * request, 
-					  char * key, 
+					  const char * key, 
 					  const void *value );
 
 void gftp_set_bookmark_option 		( gftp_bookmarks_var * bm,
-					  char * key, 
+					  const char * key, 
 					  const void *value );
 
 void gftp_register_config_vars 		( gftp_config_vars *config_vars );
 
 void gftp_copy_local_options 		( gftp_config_vars ** new_options_vars, 
 					  GHashTable ** new_options_hash,
+					  int *new_num_local_options_vars,
 					  gftp_config_vars * orig_options,
 					  int num_local_options_vars );
 
@@ -662,15 +694,10 @@ char *alltrim 				( char *str );
 
 char *expand_path 			( const char *src );
 
-void remove_double_slashes 		( char *string );
-
 void make_nonnull 			( char **str );
 
-int copyfile				( char *source,
-					  char *dest );
-
-int gftp_match_filespec 		( char *filename, 
-					  char *filespec );
+int gftp_match_filespec 		( const char *filename, 
+					  const char *filespec );
 
 int gftp_parse_command_line 		( int *argc,
 					  char ***argv );
@@ -696,8 +723,7 @@ gftp_transfer * gftp_tdata_new 		( void );
 
 void free_tdata 			( gftp_transfer * tdata );
 
-gftp_request * copy_request 		( gftp_request * req,
-					  int copy_local_options );
+gftp_request * gftp_copy_request 	( gftp_request * req );
 
 GList * gftp_sort_filelist 		( GList * filelist, 
 					  int column, 
@@ -724,13 +750,27 @@ char * gftp_build_path 			( const char *first_element,
 
 off_t gftp_parse_file_size 		( char *str );
 
+void gftp_locale_init 			( void );
+
+char * gftp_scramble_password		( const char *password );
+
+char * gftp_descramble_password		( const char *password );
+
+int gftp_get_transfer_action 		( gftp_request * request,
+					  gftp_file * fle );
+
+/* mkstemps.c */
+int mkstemps 				( char *template,
+					  int suffix_len );
+
 /* protocols.c */
 #define GFTP_FTP_NUM				0
-#define GFTP_HTTP_NUM				1
-#define GFTP_HTTPS_NUM				2
-#define GFTP_LOCAL_NUM				3
-#define GFTP_SSHV2_NUM				4
-#define GFTP_BOOKMARK_NUM			5
+#define GFTP_FTPS_NUM				1
+#define GFTP_HTTP_NUM				2
+#define GFTP_HTTPS_NUM				3
+#define GFTP_LOCAL_NUM				4
+#define GFTP_SSHV2_NUM				5
+#define GFTP_BOOKMARK_NUM			6
 
 #define GFTP_IS_CONNECTED(request)		((request) != NULL && \
                                                  ((request)->datafd > 0 || \
@@ -746,6 +786,12 @@ int rfc959_get_next_file 		( gftp_request * request,
 					  gftp_file *fle, 
 					  int fd );
 
+int rfc959_connect 			( gftp_request * request );
+
+int ftps_init 				( gftp_request * request );
+
+void ftps_register_module		( void );
+
 int rfc2068_init 			( gftp_request * request );
 
 void rfc2068_register_module		( void );
@@ -773,6 +819,9 @@ gftp_request *gftp_request_new 		( void );
 void gftp_request_destroy 		( gftp_request * request,
 					  int free_request );
 
+void gftp_copy_param_options		( gftp_request * dest_request,
+					  gftp_request * src_request );
+
 void gftp_file_destroy 			( gftp_file *file );
 
 int gftp_connect 			( gftp_request * request );
@@ -809,20 +858,22 @@ ssize_t gftp_put_next_file_chunk 	( gftp_request * request,
 
 int gftp_list_files 			( gftp_request * request );
 
-#if GLIB_MAJOR_VERSION > 1
 char * gftp_string_to_utf8		( gftp_request * request, 
-					  char *str );
-#endif
+					  const char *str );
+
+char * gftp_string_from_utf8		( gftp_request * request, 
+					  const char *str );
 
 int gftp_parse_bookmark 		( gftp_request * request, 
 					  gftp_request * local_request,
-					  const char * bookmark );
+					  const char * bookmark,
+					  int *refresh_local );
 
 int gftp_parse_url 			( gftp_request * request, 
 					  const char *url );
 
 int gftp_get_next_file 			( gftp_request * request, 
-					  char *filespec,
+					  const char *filespec,
 					  gftp_file *fle );
 
 int gftp_end_transfer 			( gftp_request * request );
@@ -951,6 +1002,8 @@ int gftp_fd_open 			( gftp_request * request,
 					  int flags,
 					  mode_t perms );
 
+void gftp_setup_startup_directory 	( gftp_request * request );
+
 /* pty.c */
 char * gftp_get_pty_impl 		( void );
 
