@@ -18,7 +18,7 @@
 /*****************************************************************************/
 
 #include "gftp.h"
-static const char cvsid[] = "$Id: local.c,v 1.26 2003/08/12 01:05:02 masneyb Exp $";
+static const char cvsid[] = "$Id: local.c,v 1.31 2004/03/18 12:01:57 masneyb Exp $";
 
 typedef struct local_protocol_data_tag
 {
@@ -116,7 +116,7 @@ local_get_file (gftp_request * request, const char *filename, int fd,
   if (fd <= 0)
     {
       flags = O_RDONLY;
-#if defined (_LARGEFILE_SOURCE)
+#if defined (_LARGEFILE_SOURCE) && defined (O_LARGEFILE)
       flags |= O_LARGEFILE;
 #endif
 
@@ -163,7 +163,7 @@ local_put_file (gftp_request * request, const char *filename, int fd,
       flags = O_WRONLY | O_CREAT;
       if (startsize > 0)
          flags |= O_APPEND;
-#if defined (_LARGEFILE_SOURCE)
+#if defined (_LARGEFILE_SOURCE) && defined (O_LARGEFILE)
       flags |= O_LARGEFILE;
 #endif
 
@@ -300,7 +300,7 @@ local_get_next_file (gftp_request * request, gftp_file * fle, int fd)
 
   /* the struct passwd and struct group are not thread safe. But,
      we're ok here because I have threading turned off for the local
-     protocol (see use_threads in local_init above) */
+     protocol (see use_threads in gftp_protocols in options.h) */
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->protonum == GFTP_LOCAL_NUM, GFTP_EFATAL);
   g_return_val_if_fail (fle != NULL, GFTP_EFATAL);
@@ -315,23 +315,15 @@ local_get_next_file (gftp_request * request, gftp_file * fle, int fd)
     {
       closedir (lpd->dir);
       lpd->dir = NULL;
-      return (GFTP_ERETRYABLE);
+      return (GFTP_EFATAL);
     }
 
   fle->file = g_strdup (dirp->d_name);
   if (lstat (fle->file, &st) != 0)
-    {
-      closedir (lpd->dir);
-      lpd->dir = NULL;
-      return (GFTP_ERETRYABLE);
-    }
+    return (GFTP_ERETRYABLE);
 
   if (stat (fle->file, &fst) != 0)
-    {
-      closedir (lpd->dir);
-      lpd->dir = NULL;
-      return (GFTP_ERETRYABLE);
-    }
+    return (GFTP_ERETRYABLE);
 
   if ((user = g_hash_table_lookup (lpd->userhash, 
                                    GUINT_TO_POINTER(st.st_uid))) != NULL)
@@ -361,14 +353,14 @@ local_get_next_file (gftp_request * request, gftp_file * fle, int fd)
       g_hash_table_insert (lpd->grouphash, GUINT_TO_POINTER (st.st_gid), group);
     }
 
-  fle->attribs = make_text_mode (st.st_mode);
+  fle->attribs = make_text_mode (fst.st_mode);
   fle->datetime = st.st_mtime;
 
   if ((fle->attribs[0] == 'b' || fle->attribs[0] == 'u' ||
        fle->attribs[0] == 'c'))
     fle->size = (off_t) st.st_rdev;
   else
-    fle->size = st.st_size;
+    fle->size = fst.st_size;
 
   fle->isdir = S_ISDIR (fst.st_mode);
   fle->islink = S_ISLNK (st.st_mode);
@@ -646,6 +638,7 @@ local_init (gftp_request * request)
 
   request->protonum = GFTP_LOCAL_NUM;
   request->init = local_init;
+  request->copy_param_options = NULL;
   request->destroy = local_destroy;
   request->read_function = gftp_fd_read;
   request->write_function = gftp_fd_write;
@@ -677,7 +670,6 @@ local_init (gftp_request * request)
   request->need_hostport = 0;
   request->need_userpass = 0;
   request->use_cache = 0;
-  request->use_threads = 0;
   request->always_connected = 1;
 
   lpd = g_malloc0 (sizeof (*lpd));
