@@ -17,12 +17,38 @@
 /*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
 /*****************************************************************************/
 
-static const char cvsid[] = "$Id: misc.c,v 1.46 2003/08/18 00:57:57 masneyb Exp $";
+static const char cvsid[] = "$Id: misc.c,v 1.66 2004/03/26 21:03:49 masneyb Exp $";
 
 #include "gftp.h"
 #include "options.h"
 
-/* FIXME - this isn't right for all locales. Use glib's printf instead */
+#ifdef HAVE_INTL_PRINTF
+
+char *
+insert_commas (off_t number, char *dest_str, size_t dest_len)
+{
+  if (dest_str != NULL)
+    {
+#if defined (_LARGEFILE_SOURCE)
+      g_snprintf (dest_str, dest_len, "%'lld", (long long) number);
+#else
+      g_snprintf (dest_str, dest_len, "%'ld", number);
+#endif
+    }
+  else
+    {
+#if defined (_LARGEFILE_SOURCE)
+      dest_str = g_strdup_printf ("%'lld", (long long) number);
+#else
+      dest_str = g_strdup_printf ("%'ld", number);
+#endif
+    }
+
+  return (dest_str);
+}
+
+#else
+
 char *
 insert_commas (off_t number, char *dest_str, size_t dest_len)
 {
@@ -30,7 +56,7 @@ insert_commas (off_t number, char *dest_str, size_t dest_len)
   int len, num, rem, i;
 
 #if defined (_LARGEFILE_SOURCE)
-  g_snprintf (src, sizeof (src), "%lld", number);
+  g_snprintf (src, sizeof (src), "%lld", (long long) number);
 #else
   g_snprintf (src, sizeof (src), "%ld", number);
 #endif
@@ -92,6 +118,7 @@ insert_commas (off_t number, char *dest_str, size_t dest_len)
   return (dest);
 }
 
+#endif
 
 char *
 alltrim (char *str)
@@ -124,7 +151,8 @@ alltrim (char *str)
 char *
 expand_path (const char *src)
 {
-  char *str, *pos, *endpos, *prevpos, *newstr, *tempstr, tempchar;
+  char *str, *pos, *endpos, *prevpos, *newstr, *tempstr, *ntoken,
+       tempchar;
   struct passwd *pw;
 
   pw = NULL;
@@ -151,22 +179,28 @@ expand_path (const char *src)
   while ((pos = strchr (endpos, '/')) != NULL)
     {
       pos++;
-      while (*pos == '/')
-        pos++;
 
-      if ((endpos = strchr (pos, '/')) == NULL)
+      for (ntoken = pos; *ntoken == '/'; ntoken++);
+
+      if ((endpos = strchr (ntoken, '/')) == NULL)
 	endpos = pos + strlen (pos);
 
       tempchar = *endpos;
       *endpos = '\0';
 
-      if (strcmp (pos, "..") == 0)
+      if (strcmp (ntoken, "..") == 0)
 	{
-	  *(pos - 1) = '\0';
 	  if (newstr != NULL && (prevpos = strrchr (newstr, '/')) != NULL)
-	    *prevpos = '\0';
+            {
+	      *prevpos = '\0';
+              if (*newstr == '\0')
+                {
+                  g_free (newstr);
+                  newstr = NULL;
+                }
+            }
 	}
-      else if (strcmp (pos, ".") != 0)
+      else if (strcmp (ntoken, ".") != 0)
 	{
 	  if (newstr == NULL)
 	    newstr = g_strdup (pos - 1);
@@ -188,7 +222,7 @@ expand_path (const char *src)
   if (endpos != NULL && *endpos != '\0' && newstr == NULL)
     {
       if (strcmp (endpos, "..") == 0)
-        newstr = g_malloc0 (1);
+        newstr = g_strdup ("/");
       else
         newstr = g_strdup (endpos);
     }
@@ -218,29 +252,6 @@ expand_path (const char *src)
 }
 
 
-void
-remove_double_slashes (char *string)
-{
-  char *newpos, *oldpos;
-  size_t len;
-
-  oldpos = newpos = string;
-  while (*oldpos != '\0')
-    {
-      *newpos++ = *oldpos++;
-      if (*oldpos == '\0')
-	break;
-      while (*(newpos - 1) == '/' && *(oldpos) == '/')
-	oldpos++;
-    }
-  *newpos = '\0';
-
-  len = strlen (string);
-  if (string[len - 1] == '/')
-    string[len - 1] = '\0';
-}
-
-
 void
 make_nonnull (char **str)
 {
@@ -249,65 +260,12 @@ make_nonnull (char **str)
 }
 
 
-int
-copyfile (char *source, char *dest)
-{
-  int srcfd, destfd;
-  char buf[8192];
-  ssize_t n;
-
-  if ((srcfd = gftp_fd_open (NULL, source, O_RDONLY, 0)) == -1)
-    {
-      printf (_("Error: Cannot open local file %s: %s\n"),
-              source, g_strerror (errno));
-      exit (1);
-    }
-
-  if ((destfd = gftp_fd_open (NULL, dest, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR)) == -1)
-    {
-      printf (_("Error: Cannot open local file %s: %s\n"),
-              dest, g_strerror (errno));
-      close (srcfd);
-      exit (1);
-    }
-
-  while ((n = read (srcfd, buf, sizeof (buf))) > 0)
-    {
-      if (write (destfd, buf, n) == -1)
-        {
-          printf (_("Error: Could not write to socket: %s\n"), 
-                  g_strerror (errno));
-          exit (1);
-        }
-    }
-
-  if (n == -1)
-    {
-      printf (_("Error: Could not read from socket: %s\n"), g_strerror (errno));
-      exit (1);
-    }
-
-  if (close (srcfd) == -1)
-    {
-      printf (_("Error closing file descriptor: %s\n"), g_strerror (errno));
-      exit (1);
-    }
-
-  if (close (destfd) == -1)
-    {
-      printf (_("Error closing file descriptor: %s\n"), g_strerror (errno));
-      exit (1);
-    }
-
-  return (1);
-}
-
-
 /* FIXME - is there a replacement for this */
 int
-gftp_match_filespec (char *filename, char *filespec)
+gftp_match_filespec (const char *filename, const char *filespec)
 {
-  char *filepos, *wcpos, *pos, *newpos, search_str[20];
+  const char *filepos, *wcpos, *pos;
+  char search_str[20], *newpos;
   size_t len, curlen;
   
   if (filename == NULL || *filename == '\0' || 
@@ -356,6 +314,14 @@ gftp_info (void)
 
   printf ("%s\n", gftp_version);
 
+#ifdef _REENTRANT
+  printf ("#define _REENTRANT\n");
+#endif
+
+#ifdef _GNU_SOURCE
+  printf ("#define _GNU_SOURCE\n");
+#endif
+
 #ifdef _LARGEFILE_SOURCE
   printf ("#define _LARGEFILE_SOURCE\n");
 #endif
@@ -394,6 +360,10 @@ gftp_info (void)
   printf ("#define HAVE_GETTEXT\n");
 #endif
 
+#ifdef HAVE_INTL_PRINTF
+  printf ("#define HAVE_INTL_PRINTF\n");
+#endif
+
   printf ("glib version: %d.%d.%d\n", GLIB_MAJOR_VERSION, GLIB_MINOR_VERSION,
           GLIB_MICRO_VERSION);
 
@@ -419,17 +389,20 @@ gftp_parse_command_line (int *argc, char ***argv)
     {
       if (strcmp (argv[0][1], "--help") == 0 || 
           strcmp (argv[0][1], "-h") == 0)
-	gftp_usage ();
+        {
+          gftp_usage ();
+          exit (0);
+        }
       else if (strcmp (argv[0][1], "--version") == 0 || 
                strcmp (argv[0][1], "-v") == 0)
 	{
-	  printf ("%s\n", gftp_version);
-	  exit (0);
+          printf ("%s\n", gftp_version);
+          exit (0);
 	}
       else if (strcmp (argv[0][1], "--info") == 0)
 	{
           gftp_info ();
-	  exit (0);
+          exit (0);
 	}
     }
   return (0);
@@ -439,7 +412,7 @@ gftp_parse_command_line (int *argc, char ***argv)
 void
 gftp_usage (void)
 {
-  printf (_("usage: gftp [[protocol://][user[:pass]@]site[:port][/directory]]\n"));
+  printf (_("usage: gftp " GFTP_URL_USAGE "\n"));
   exit (0);
 }
 
@@ -614,7 +587,7 @@ free_tdata (gftp_transfer * tdata)
 
 
 gftp_request * 
-copy_request (gftp_request * req, int copy_local_options)
+gftp_copy_request (gftp_request * req)
 {
   gftp_request * newreq;
 
@@ -634,20 +607,22 @@ copy_request (gftp_request * req, int copy_local_options)
   newreq->use_proxy = req->use_proxy;
   newreq->logging_function = req->logging_function;
   newreq->free_hostp = 0;
-  newreq->hostp = req->hostp;
+  newreq->hostp = NULL;
 
-  if (copy_local_options)
-    gftp_copy_local_options (&newreq->local_options_vars, 
-                             &newreq->local_options_hash,
-                             req->local_options_vars,
-                             req->num_local_options_vars);
+  gftp_copy_local_options (&newreq->local_options_vars, 
+                           &newreq->local_options_hash,
+                           &newreq->num_local_options_vars,
+                           req->local_options_vars,
+                           req->num_local_options_vars);
 
-  if (req->init (newreq) < 0)
+  if (req->init != NULL && req->init (newreq) < 0)
     {
       gftp_request_destroy (newreq, 1);
       return (NULL);
     }
 
+  gftp_copy_param_options (newreq, req);
+
   return (newreq);
 }
 
@@ -818,7 +793,10 @@ gftp_sort_filelist (GList * filelist, int column, int asds)
   GList * files, * dirs, * dotdot, * tempitem, * insitem;
   GCompareFunc sortfunc;
   gftp_file * tempfle;
-  int sort_dirs_first;
+  intptr_t sort_dirs_first;
+
+  if (filelist == NULL) /* nothing to sort */
+    return (filelist);
 
   files = dirs = dotdot = NULL;
 
@@ -945,7 +923,7 @@ gftp_gen_ls_string (gftp_file * fle, char *file_prefixstr, char *file_suffixstr)
   else
     {
 #if defined (_LARGEFILE_SOURCE)
-      tempstr2 = g_strdup_printf ("%11lld", fle->size);
+      tempstr2 = g_strdup_printf ("%11lld", (long long) fle->size);
 #else
       tempstr2 = g_strdup_printf ("%11ld", fle->size);
 #endif
@@ -1221,7 +1199,10 @@ gftp_build_path (const char *first_element, ...)
     {
       len = strlen (element);
 
-      if (len > 0 && element[len - 1] == '/')
+      if (len == 0)
+        continue;
+
+      if (retlen > 0 && (ret[retlen - 1] == '/' || element[0] == '/'))
         add_separator = 0;
       else
         {
@@ -1252,3 +1233,113 @@ gftp_parse_file_size (char *str)
 #endif
 }
 
+
+void
+gftp_locale_init (void)
+{
+#ifdef HAVE_GETTEXT
+
+  setlocale (LC_ALL, "");
+  textdomain ("gftp");
+  bindtextdomain ("gftp", LOCALE_DIR);
+
+#if GLIB_MAJOR_VERSION > 1
+  bind_textdomain_codeset ("gftp", "UTF-8");
+#endif
+
+#endif /* HAVE_GETTEXT */
+}
+
+/* Very primary encryption/decryption to make the passwords unreadable
+   with 'cat ~/.gftp/bookmarks'.
+   
+   Each character is separated in two nibbles. Then each nibble is stored
+   under the form 01xxxx01. The resulted string is prefixed by a '$'.
+*/
+
+
+char * 
+gftp_scramble_password (const char *password)
+{
+  char *newstr, *newpos;
+
+  if (strcmp (password, "@EMAIL@") == 0)
+    return (g_strdup (password));
+
+  newstr = g_malloc (strlen(password) * 2 + 2);
+  newpos = newstr;
+  
+  *newpos++ = '$';
+
+  while (*password != 0)
+    {
+      *newpos++ = ((*password >> 2) & 0x3c) | 0x41;
+      *newpos++ = ((*password << 2) & 0x3c) | 0x41;
+      password++;
+    }
+  *newpos = 0;
+  
+  return (newstr);
+}
+
+
+char *
+gftp_descramble_password (const char *password)
+{
+  const char *passwordpos;
+  char *newstr, *newpos;
+  int error;
+
+  if (*password != '$')
+    return (g_strdup (password));
+
+  passwordpos = password + 1;
+  newstr = g_malloc (strlen (passwordpos) / 2 + 1);
+  newpos = newstr;
+ 
+  error = 0;
+  while (*passwordpos != '\0' && (*passwordpos + 1) != '\0')
+    {
+      if ((*passwordpos & 0xc3) != 0x41 ||
+          (*(passwordpos + 1) & 0xc3) != 0x41)
+        {
+          error = 1;
+          break;
+        }
+
+      *newpos++ = ((*passwordpos & 0x3c) << 2) | 
+                  ((*(passwordpos + 1) & 0x3c) >> 2);
+
+      passwordpos += 2;
+    }
+
+  if (error)
+    {
+      g_free (newstr);
+      return (g_strdup (password));
+    }
+
+  *newpos = '\0';
+  return (newstr);
+}
+
+
+int
+gftp_get_transfer_action (gftp_request * request, gftp_file * fle)
+{
+  intptr_t overwrite_default;
+
+  gftp_lookup_request_option (request, "overwrite_default", &overwrite_default);
+
+  if (overwrite_default)
+    fle->transfer_action = GFTP_TRANS_ACTION_OVERWRITE;
+  else if (fle->startsize == fle->size)
+    fle->transfer_action = GFTP_TRANS_ACTION_SKIP;
+  else if (fle->startsize > fle->size)
+    fle->transfer_action = GFTP_TRANS_ACTION_OVERWRITE;
+  else
+    fle->transfer_action = GFTP_TRANS_ACTION_RESUME;
+
+  return (fle->transfer_action);
+}
+
