@@ -17,7 +17,7 @@
 /*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
 /*****************************************************************************/
 
-/* $Id: options.h,v 1.20 2003/07/21 00:26:41 masneyb Exp $ */
+/* $Id: options.h,v 1.28 2004/03/17 20:42:52 masneyb Exp $ */
 
 #include "gftp.h"
 
@@ -85,6 +85,16 @@ gftp_config_vars gftp_global_config_vars[] =
    gftp_option_type_checkbox, GINT_TO_POINTER(1), NULL, 
    GFTP_CVARS_FLAGS_SHOW_BOOKMARK,
    N_("Show hidden files in the listboxes"), GFTP_PORT_ALL, NULL},
+  {"show_trans_in_title", N_("Show transfer status in title"), 
+   gftp_option_type_checkbox, GINT_TO_POINTER(0), NULL, 0,
+   N_("Show the file transfer status in the titlebar"), GFTP_PORT_GTK, NULL},
+  {"start_transfers", N_("Start file transfers"), 
+   gftp_option_type_checkbox, GINT_TO_POINTER(1), NULL, 0,
+   N_("Automatically start the file transfers when they get queued"),
+   GFTP_PORT_GTK, NULL},
+  {"cmd_in_gui", N_("Allow manual commands in GUI"), 
+   gftp_option_type_checkbox, GINT_TO_POINTER(0), NULL, 0,
+   N_("Allow entering manual commands in the GUI (functions like the text port)"), GFTP_PORT_GTK, NULL},
 
   {"", N_("Network"), gftp_option_type_notebook, NULL, NULL, 
    GFTP_CVARS_FLAGS_SHOW_BOOKMARK, NULL, GFTP_PORT_GTK, NULL},
@@ -111,6 +121,12 @@ gftp_config_vars gftp_global_config_vars[] =
   {"default_protocol", N_("Default Protocol:"),
    gftp_option_type_textcombo, "FTP", NULL, 0,
    N_("This specifies the default protocol to use"), GFTP_PORT_ALL, NULL},
+#if defined (HAVE_GETADDRINFO) && defined (HAVE_GAI_STRERROR)
+  {"enable_ipv6", N_("Enable IPv6 support"), 
+   gftp_option_type_checkbox, GINT_TO_POINTER(1), NULL, 
+   GFTP_CVARS_FLAGS_SHOW_BOOKMARK,
+   N_("Enable IPv6 support"), GFTP_PORT_ALL, NULL},
+#endif
 
   {"list_dblclk_action", "", 
    gftp_option_type_int, GINT_TO_POINTER(0), NULL, 0,
@@ -200,16 +216,28 @@ gftp_config_vars gftp_global_config_vars[] =
 
 supported_gftp_protocols gftp_protocols[] =
 {
-  {N_("FTP"), rfc959_init, rfc959_register_module, "ftp", 1},
-  {N_("HTTP"), rfc2068_init, rfc2068_register_module, "http", 1},
+  {N_("FTP"), rfc959_init, rfc959_register_module, "ftp", 1, 1},
+
 #ifdef USE_SSL
-  {N_("HTTPS"), https_init, https_register_module, "https", 1},
+  {N_("FTPS"), ftps_init, ftps_register_module, "ftps", 1, 1},
 #else
-  {N_("HTTPS"), https_init, https_register_module, "https", 0},
+  {N_("FTPS"), ftps_init, ftps_register_module, "ftps", 0, 1},
 #endif
-  {N_("Local"), local_init, local_register_module, "file", 1},
-  {N_("SSH2"), sshv2_init, sshv2_register_module, "ssh2", 1},
-  {N_("Bookmark"), bookmark_init, bookmark_register_module, "bookmark", 0},
+
+  {N_("HTTP"), rfc2068_init, rfc2068_register_module, "http", 1, 1},
+
+#ifdef USE_SSL
+  {N_("HTTPS"), https_init, https_register_module, "https", 1, 1},
+#else
+  {N_("HTTPS"), https_init, https_register_module, "https", 0, 1},
+#endif
+
+  {N_("Local"), local_init, local_register_module, "file", 1, 0},
+
+  {N_("SSH2"), sshv2_init, sshv2_register_module, "ssh2", 1, 1},
+
+  {N_("Bookmark"), bookmark_init, bookmark_register_module, "bookmark", 0, 0},
+
   {NULL, NULL, NULL, NULL, 0}
 };
 
