@@ -18,7 +18,7 @@
 /*****************************************************************************/
 
 #include "gftp.h"
-static const char cvsid[] = "$Id: protocols.c,v 1.64 2003/08/21 20:24:08 masneyb Exp $";
+static const char cvsid[] = "$Id: protocols.c,v 1.86 2004/03/26 21:03:49 masneyb Exp $";
 
 gftp_request *
 gftp_request_new (void)
@@ -83,6 +83,22 @@ gftp_request_destroy (gftp_request * request, int free_request)
 }
 
 
+/* This function is called to copy protocol specific data from one request 
+   structure to another. This is typically called when a file transfer is
+   completed, state information can be copied back to the main window */
+void
+gftp_copy_param_options (gftp_request * dest_request,
+                         gftp_request * src_request)
+{
+  g_return_if_fail (dest_request != NULL);
+  g_return_if_fail (src_request != NULL);
+  g_return_if_fail (dest_request->protonum == src_request->protonum);
+
+  if (dest_request->copy_param_options)
+    dest_request->copy_param_options (dest_request, src_request);
+}
+
+
 void
 gftp_file_destroy (gftp_file * file)
 {
@@ -90,6 +106,8 @@ gftp_file_destroy (gftp_file * file)
 
   if (file->file)
     g_free (file->file);
+  if (file->utf8_file)
+    g_free (file->utf8_file);
   if (file->user)
     g_free (file->user);
   if (file->group)
@@ -157,18 +175,8 @@ off_t
 gftp_get_file (gftp_request * request, const char *filename, int fd,
                off_t startsize)
 {
-  float maxkbs;
-
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
 
-  gftp_lookup_request_option (request, "maxkbs", &maxkbs);
-  if (maxkbs > 0)
-    {
-      request->logging_function (gftp_logging_misc, request,
-                    _("File transfer will be throttled to %.2f KB/s\n"),
-                    maxkbs);
-    }
-
   request->cached = 0;
   if (request->get_file == NULL)
     return (GFTP_EFATAL);
@@ -181,19 +189,8 @@ int
 gftp_put_file (gftp_request * request, const char *filename, int fd,
                off_t startsize, off_t totalsize)
 {
-  float maxkbs;
-
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
 
-  gftp_lookup_request_option (request, "maxkbs", &maxkbs);
-
-  if (maxkbs > 0)
-    {
-      request->logging_function (gftp_logging_misc, request,
-                    _("File transfer will be throttled to %.2f KB/s\n"), 
-                    maxkbs);
-    }
-
   request->cached = 0;
   if (request->put_file == NULL)
     return (GFTP_EFATAL);
@@ -207,6 +204,7 @@ gftp_transfer_file (gftp_request * fromreq, const char *fromfile,
                     gftp_request * toreq, const char *tofile,
                     int tofd, off_t tosize)
 {
+  float maxkbs;
   off_t size;
   int ret;
 
@@ -215,6 +213,15 @@ gftp_transfer_file (gftp_request * fromreq, const char *fromfile,
   g_return_val_if_fail (toreq != NULL, GFTP_EFATAL);
   g_return_val_if_fail (tofile != NULL, GFTP_EFATAL);
 
+  gftp_lookup_request_option (toreq, "maxkbs", &maxkbs);
+
+  if (maxkbs > 0)
+    {
+      toreq->logging_function (gftp_logging_misc, toreq,
+                    _("File transfer will be throttled to %.2f KB/s\n"), 
+                    maxkbs);
+    }
+
   if (fromreq->protonum == toreq->protonum &&
       fromreq->transfer_file != NULL)
     return (fromreq->transfer_file (fromreq, fromfile, fromsize, toreq, 
@@ -222,11 +229,36 @@ gftp_transfer_file (gftp_request * fromreq, const char *fromfile,
 
   fromreq->cached = 0;
   toreq->cached = 0;
-  if ((size = gftp_get_file (fromreq, fromfile, fromfd, tosize)) < 0)
-    return (size);
 
-  if ((ret = gftp_put_file (toreq, tofile, tofd, tosize, size)) != 0)
+get_file:
+  size = gftp_get_file (fromreq, fromfile, fromfd, tosize);
+  if (size < 0)
     {
+      if (size == GFTP_ETIMEDOUT)
+        {
+          ret = gftp_connect (fromreq);
+          if (ret < 0)
+            return (ret);
+
+          goto get_file;
+        }
+
+      return (size);
+    }
+
+put_file:
+  ret = gftp_put_file (toreq, tofile, tofd, tosize, size);
+  if (ret != 0)
+    {
+      if (size == GFTP_ETIMEDOUT)
+        {
+          ret = gftp_connect (fromreq);
+          if (ret < 0)
+            return (ret);
+
+          goto put_file;
+        }
+
       if (gftp_abort_transfer (fromreq) != 0)
         gftp_end_transfer (fromreq);
 
@@ -339,8 +371,9 @@ gftp_list_files (gftp_request * request)
 
 
 #if GLIB_MAJOR_VERSION > 1
+
 static char *
-_gftp_get_next_charset (char *remote_charsets, char *orig_str, char **curpos)
+_gftp_get_next_charset (char **curpos)
 {
   char *ret, *endpos;
 
@@ -348,40 +381,22 @@ _gftp_get_next_charset (char *remote_charsets, char *orig_str, char **curpos)
     return (NULL);
 
   ret = *curpos;
-  if (*curpos != remote_charsets)
-    {
-      *orig_str = *(*curpos - 1);
-      *(*curpos - 1) = ',';
-    }
-
   if ((endpos = strchr (*curpos, ',')) == NULL)
     *curpos += strlen (*curpos);
   else
     {
       *endpos = '\0';
-
-      if (*orig_str != '\0')
-        *curpos = endpos + 1;
-      else
-        *curpos = endpos;
+      *curpos = endpos + 1;
     }
 
   return (ret);
 }
 
 
-static void
-_gftp_restore_charset_string (char **remote_charsets, char orig_str, char **curpos)
-{
-  if (*curpos != *remote_charsets)
-    *(*curpos - 1) = orig_str;
-}
-
-
 char *
-gftp_string_to_utf8 (gftp_request * request, char *str)
+gftp_string_to_utf8 (gftp_request * request, const char *str)
 {
-  char *ret, *remote_charsets, *stpos, *cur_charset, orig_str;
+  char *ret, *remote_charsets, *stpos, *cur_charset, *tempstr;
   gsize bread, bwrite;
   GError * error;
 
@@ -391,9 +406,11 @@ gftp_string_to_utf8 (gftp_request * request, char *str)
   if (request->iconv_initialized)
     return (g_convert_with_iconv (str, -1, request->iconv, &bread, &bwrite, 
                                   &error));
+  else if (g_utf8_validate (str, -1, NULL))
+    return (NULL);
 
-  gftp_lookup_request_option (request, "remote_charsets", &remote_charsets);
-  if (*remote_charsets == '\0')
+  gftp_lookup_request_option (request, "remote_charsets", &tempstr);
+  if (*tempstr == '\0')
     {
       error = NULL;
       if ((ret = g_locale_to_utf8 (str, -1, &bread, &bwrite, &error)) != NULL)
@@ -402,10 +419,10 @@ gftp_string_to_utf8 (gftp_request * request, char *str)
       return (NULL);
     }
 
+  remote_charsets = g_strdup (tempstr);
   ret = NULL;
   stpos = remote_charsets;
-  while ((cur_charset = _gftp_get_next_charset (remote_charsets, &orig_str,
-                                                &stpos)) != NULL)
+  while ((cur_charset = _gftp_get_next_charset (&stpos)) != NULL)
     {
       if ((request->iconv = g_iconv_open ("UTF-8", cur_charset)) == (GIConv) -1)
         continue;
@@ -421,18 +438,91 @@ gftp_string_to_utf8 (gftp_request * request, char *str)
       else
         {
           request->iconv_initialized = 1;
-          _gftp_restore_charset_string (&remote_charsets, *cur_charset, &stpos);
           break;
         }
     }
 
+  g_free (remote_charsets);
+
   return (ret);
 }
+
+
+char *
+gftp_string_from_utf8 (gftp_request * request, const char *str)
+{
+  char *ret, *remote_charsets, *stpos, *cur_charset, *tempstr;
+  gsize bread, bwrite;
+  GError * error;
+
+  if (request == NULL)
+    return (NULL);
+
+  if (request->iconv_initialized)
+    return (g_convert_with_iconv (str, -1, request->iconv, &bread, &bwrite, 
+                                  &error));
+  else if (g_utf8_validate (str, -1, NULL))
+    return (NULL);
+
+  gftp_lookup_request_option (request, "remote_charsets", &tempstr);
+  if (*tempstr == '\0')
+    {
+      error = NULL;
+      if ((ret = g_locale_from_utf8 (str, -1, &bread, &bwrite, &error)) != NULL)
+        return (ret);
+
+      return (NULL);
+    }
+
+  remote_charsets = g_strdup (tempstr);
+  ret = NULL;
+  stpos = remote_charsets;
+  while ((cur_charset = _gftp_get_next_charset (&stpos)) != NULL)
+    {
+      if ((request->iconv = g_iconv_open (cur_charset, "UTF-8")) == (GIConv) -1)
+        continue;
+
+      error = NULL;
+      if ((ret = g_convert_with_iconv (str, -1, request->iconv, &bread, &bwrite,
+                                       &error)) == NULL)
+        {
+          g_iconv_close (request->iconv);
+          request->iconv = NULL;
+          continue;
+        }
+      else
+        {
+          request->iconv_initialized = 1;
+          break;
+        }
+    }
+
+  g_free (remote_charsets);
+
+  return (ret);
+}
+
+#else
+
+char *
+gftp_string_to_utf8 (gftp_request * request, const char *str)
+{
+  return (NULL);
+}
+
+
+char *
+gftp_string_from_utf8 (gftp_request * request, const char *str)
+{
+  return (NULL);
+}
+
 #endif
 
 
 int
-gftp_get_next_file (gftp_request * request, char *filespec, gftp_file * fle)
+gftp_get_next_file (gftp_request * request, const char *filespec,
+                    gftp_file * fle)
 {
   int fd, ret;
 
@@ -452,10 +542,8 @@ gftp_get_next_file (gftp_request * request, char *filespec, gftp_file * fle)
       gftp_file_destroy (fle);
       ret = request->get_next_file (request, fle, fd);
 
-#if GLIB_MAJOR_VERSION > 1
-      if (ret >= 0 && fle->file != NULL && !g_utf8_validate (fle->file, -1, NULL))
+      if (ret >= 0 && fle->file != NULL)
         fle->utf8_file = gftp_string_to_utf8 (request, fle->file);
-#endif
 
       if (ret >= 0 && !request->cached && request->cachefd > 0 && 
           request->last_dir_entry != NULL)
@@ -478,7 +566,7 @@ gftp_get_next_file (gftp_request * request, char *filespec, gftp_file * fle)
 
 int
 gftp_parse_bookmark (gftp_request * request, gftp_request * local_request, 
-                     const char * bookmark)
+                     const char * bookmark, int *refresh_local)
 {
   gftp_logging_func logging_function;
   gftp_bookmarks_var * tempentry;
@@ -523,7 +611,12 @@ gftp_parse_bookmark (gftp_request * request, gftp_request * local_request,
   if (local_request != NULL &&
       tempentry->local_dir != NULL &&
       *tempentry->local_dir != '\0')
-    gftp_set_directory (local_request, tempentry->local_dir);
+    {
+      gftp_set_directory (local_request, tempentry->local_dir);
+      *refresh_local = 1;
+    }
+  else
+    *refresh_local = 0;
 
   for (i = 0; gftp_protocols[i].name; i++)
     {
@@ -558,6 +651,7 @@ gftp_parse_bookmark (gftp_request * request, gftp_request * local_request,
 
   gftp_copy_local_options (&request->local_options_vars,
                            &request->local_options_hash,
+                           &request->num_local_options_vars,
                            tempentry->local_options_vars,
                            tempentry->num_local_options_vars);
 
@@ -652,6 +746,7 @@ gftp_parse_url (gftp_request * request, const char *url)
 
   if (i == GFTP_LOCAL_NUM)
     {
+      for (; *pos == ' ' || *pos == '\t'; pos++);
       request->directory = g_strdup (pos);
       g_free (stpos);
       return (0);
@@ -738,7 +833,6 @@ void
 gftp_set_password (gftp_request * request, const char *password)
 {
   g_return_if_fail (request != NULL);
-  g_return_if_fail (password != NULL);
 
   if (request->password)
     g_free (request->password);
@@ -819,11 +913,24 @@ gftp_remove_file (gftp_request * request, const char *file)
 int
 gftp_make_directory (gftp_request * request, const char *directory)
 {
+  char *utf8;
+  int ret;
+
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
 
   if (request->mkdir == NULL)
     return (GFTP_EFATAL);
-  return (request->mkdir (request, directory));
+
+  utf8 = gftp_string_from_utf8 (request, directory);
+  if (utf8 != NULL)
+    {
+      ret = request->mkdir (request, utf8);
+      g_free (utf8);
+    }
+  else
+    ret = request->mkdir (request, directory);
+
+  return (ret);
 }
 
 
@@ -831,11 +938,24 @@ int
 gftp_rename_file (gftp_request * request, const char *oldname,
                   const char *newname)
 {
+  char *utf8;
+  int ret;
+
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
 
   if (request->rename == NULL)
     return (GFTP_EFATAL);
-  return (request->rename (request, oldname, newname));
+
+  utf8 = gftp_string_from_utf8 (request, newname);
+  if (utf8 != NULL)
+    {
+      ret = request->rename (request, oldname, utf8);
+      g_free (utf8);
+    }
+  else
+    ret = request->rename (request, oldname, newname);
+
+  return (ret);
 }
 
 
@@ -1049,6 +1169,11 @@ parse_time (char *str, char **endpos)
       /* 10-Jan-2003 09:14 */
       tmppos = strptime (str, "%d-%h-%Y %H:%M", &curtime);
     }
+  else if (strlen (str) > 4 && isdigit ((int) str[0]) && str[4] == '/')
+    {
+      /* 2003/12/25 */
+      tmppos = strptime (str, "%Y/%m/%d", &curtime);
+    }
   else
     {
       /* This is how most UNIX, Novell, and MacOS ftp servers send their time */
@@ -1059,7 +1184,11 @@ parse_time (char *str, char **endpos)
           tmppos = strptime (str, "%h %d %H:%M", &curtime);
           t = time (NULL);
           loctime = localtime (&t);
-          curtime.tm_year = loctime->tm_year;
+
+          if (curtime.tm_mon > loctime->tm_mon)
+            curtime.tm_year = loctime->tm_year - 1;
+          else
+            curtime.tm_year = loctime->tm_year;
         }
       else
         tmppos = strptime (str, "%h %d %Y", &curtime);
@@ -1095,7 +1224,7 @@ parse_time (char *str, char **endpos)
         *endpos = tmppos;
     }
 
-  return (mktime (&curtime));
+  return (ret);
 }
 
 
@@ -1133,13 +1262,14 @@ gftp_parse_ls_vms (char *str, gftp_file * fle)
     return (GFTP_EFATAL);
 
   *curpos = '\0';
-  fle->attribs = g_strdup ("----------");
   if (strlen (str) > 4 && strcmp (curpos - 4, ".DIR") == 0)
     {
       fle->isdir = 1;
-      *fle->attribs = 'd';
+      fle->attribs = g_strdup ("drwxr-xr-x");
       *(curpos - 4) = '\0';
     }
+  else
+   fle->attribs = g_strdup ("-rw-r--r--");
 
   fle->file = g_strdup (str);
 
@@ -1172,22 +1302,81 @@ gftp_parse_ls_vms (char *str, gftp_file * fle)
 
   return (0);
 }
+
+
+static int
+gftp_parse_ls_mvs (char *str, gftp_file * fle)
+{
+  char *curpos;
+
+  /* Volume Unit    Referred Ext Used Recfm Lrecl BlkSz Dsorg Dsname */
+  /* SVI52A 3390   2003/12/10  8  216  FB      80 27920  PS  CARDS.DELETES */
+  /* SVI528 3390   2003/12/12  1    5  FB      80 24000  PO  CLIST */
+
+  curpos = goto_next_token (str + 1);
+  if (curpos == NULL)
+    return (GFTP_EFATAL);
+
+  curpos = goto_next_token (curpos + 1);
+  if (curpos == NULL)
+    return (GFTP_EFATAL);
+
+  if ((fle->datetime = parse_time (curpos, &curpos)) == 0)
+    return (GFTP_EFATAL);
+  curpos = goto_next_token (curpos);
+
+  curpos = goto_next_token (curpos + 1);
+  if (curpos == NULL)
+    return (GFTP_EFATAL);
+
+  fle->size = gftp_parse_file_size (curpos) * 55996; 
+  curpos = goto_next_token (curpos + 1);
+  if (curpos == NULL)
+    return (GFTP_EFATAL);
+
+  curpos = goto_next_token (curpos + 1);
+  if (curpos == NULL)
+    return (GFTP_EFATAL);
+
+  curpos = goto_next_token (curpos + 1);
+  if (curpos == NULL)
+    return (GFTP_EFATAL);
+
+  curpos = goto_next_token (curpos + 1);
+  if (curpos == NULL)
+    return (GFTP_EFATAL);
+
+  if (strncmp (curpos, "PS", 2) == 0)
+    fle->attribs = g_strdup ("-rw-r--r--");
+  else if (strncmp (curpos, "PO", 2) == 0)
+    fle->attribs = g_strdup ("drwxr-xr-x");
+  else
+    return (GFTP_EFATAL);
+
+  curpos = goto_next_token (curpos + 1);
+
+  fle->user = g_strdup (_("unknown"));
+  fle->group = g_strdup (_("unknown"));
+  fle->file = g_strdup (curpos);
+
+  return (0);
+}
   
 
 static int
 gftp_parse_ls_eplf (char *str, gftp_file * fle)
 {
   char *startpos;
+  int isdir = 0;
 
   startpos = str;
-  fle->attribs = g_strdup ("----------");
   while (startpos)
     {
       startpos++;
       switch (*startpos)
         {
         case '/':
-          *fle->attribs = 'd';
+          isdir = 1;
           break;
         case 's':
           fle->size = gftp_parse_file_size (startpos + 1);
@@ -1198,8 +1387,15 @@ gftp_parse_ls_eplf (char *str, gftp_file * fle)
         }
       startpos = strchr (startpos, ',');
     }
+
   if ((startpos = strchr (str, 9)) == NULL)
     return (GFTP_EFATAL);
+
+  if (isdir)
+    fle->attribs = g_strdup ("drwxr-xr-x");
+  else
+    fle->attribs = g_strdup ("-rw-r--r--");
+
   fle->file = g_strdup (startpos + 1);
   fle->user = g_strdup (_("unknown"));
   fle->group = g_strdup (_("unknown"));
@@ -1272,7 +1468,7 @@ gftp_parse_ls_unix (gftp_request * request, char *str, gftp_file * fle)
       startpos = goto_next_token (startpos);
     }
 
-  if (request->server_type != GFTP_DIRTYPE_CRAY)
+  if (request->server_type == GFTP_DIRTYPE_CRAY)
     {
       /* See if this is a Cray directory listing. It has the following format:
       drwx------     2 feiliu    g913     DK  common      4096 Sep 24  2001 wv */
@@ -1433,6 +1629,9 @@ gftp_parse_ls (gftp_request * request, const char *lsoutput, gftp_file * fle)
       case GFTP_DIRTYPE_VMS:
         result = gftp_parse_ls_vms (str, fle);
         break;
+      case GFTP_DIRTYPE_MVS:
+        result = gftp_parse_ls_mvs (str, fle);
+        break;
       default: /* autodetect */
         if (*lsoutput == '+')
           result = gftp_parse_ls_eplf (str, fle);
@@ -1559,15 +1758,20 @@ gftp_get_dir_listing (gftp_transfer * transfer, int getothdir)
           (newsize = g_hash_table_lookup (dirhash, fle->file)) != NULL)
         fle->startsize = *newsize;
 
-      if (transfer->toreq)
+      if (transfer->toreq && fle->destfile == NULL)
         fle->destfile = gftp_build_path (transfer->toreq->directory, 
                                          fle->file, NULL);
 
-      newname = gftp_build_path (transfer->fromreq->directory,
-                                 fle->file, NULL);
+      if (transfer->fromreq->directory != NULL &&
+          *transfer->fromreq->directory != '\0' &&
+          *fle->file != '/')
+        {
+          newname = gftp_build_path (transfer->fromreq->directory,
+                                     fle->file, NULL);
 
-      g_free (fle->file);
-      fle->file = newname;
+          g_free (fle->file);
+          fle->file = newname;
+        }
 
       templist = g_list_append (templist, fle);
 
@@ -1590,7 +1794,7 @@ int
 gftp_get_all_subdirs (gftp_transfer * transfer,
                       void (*update_func) (gftp_transfer * transfer))
 {
-  char *oldfromdir, *oldtodir, *newname;
+  char *oldfromdir, *oldtodir, *newname, *pos;
   GList * templist, * lastlist;
   int forcecd, remotechanged;
   unsigned long *newsize;
@@ -1609,16 +1813,32 @@ gftp_get_all_subdirs (gftp_transfer * transfer,
   for (lastlist = transfer->files; ; lastlist = lastlist->next)
     {
       curfle = lastlist->data;
-      if (dirhash && 
-          (newsize = g_hash_table_lookup (dirhash, curfle->file)) != NULL)
+
+      if ((pos = strrchr (curfle->file, '/')) != NULL)
+        pos++;
+      else
+        pos = curfle->file;
+
+      if (dirhash != NULL && 
+          (newsize = g_hash_table_lookup (dirhash, pos)) != NULL)
         curfle->startsize = *newsize;
 
-      if (transfer->toreq)
-        curfle->destfile = g_strconcat (transfer->toreq->directory, "/", 
-                                        curfle->file, NULL);
-      newname = g_strconcat (transfer->fromreq->directory, transfer->fromreq->directory[strlen (transfer->fromreq->directory) - 1] == '/' ? "" : "/", curfle->file, NULL);
-      g_free (curfle->file);
-      curfle->file = newname;
+      if (curfle->size < 0 && GFTP_IS_CONNECTED (transfer->fromreq))
+        curfle->size = gftp_get_file_size (transfer->fromreq, curfle->file);
+
+      if (transfer->toreq && curfle->destfile == NULL)
+        curfle->destfile = gftp_build_path (transfer->toreq->directory, 
+                                            curfle->file, NULL);
+
+      if (transfer->fromreq->directory != NULL &&
+          *transfer->fromreq->directory != '\0' &&
+          *curfle->file != '/')
+        {
+          newname = gftp_build_path (transfer->fromreq->directory,
+                                     curfle->file, NULL);
+          g_free (curfle->file);
+          curfle->file = newname;
+        }
 
       if (lastlist->next == NULL)
         break;
@@ -1720,8 +1940,8 @@ gftp_connect_server (gftp_request * request, char *service,
   int port, sock;
 #if defined (HAVE_GETADDRINFO) && defined (HAVE_GAI_STRERROR)
   struct addrinfo hints, *res;
+  int errnum, enable_ipv6;
   char serv[8];
-  int errnum;
 
   if ((request->use_proxy = gftp_need_proxy (request, service,
                                              proxy_hostname, proxy_port)) < 0)
@@ -1729,10 +1949,17 @@ gftp_connect_server (gftp_request * request, char *service,
   else if (request->use_proxy == 1)
     request->hostp = NULL;
 
+  gftp_lookup_request_option (request, "enable_ipv6", &enable_ipv6);
+
   request->free_hostp = 1;
   memset (&hints, 0, sizeof (hints));
   hints.ai_flags = AI_CANONNAME;
-  hints.ai_family = PF_UNSPEC;
+
+  if (enable_ipv6)
+    hints.ai_family = PF_UNSPEC;
+  else
+    hints.ai_family = AF_INET;
+
   hints.ai_socktype = SOCK_STREAM;
 
   if (request->use_proxy)
@@ -1820,7 +2047,7 @@ gftp_connect_server (gftp_request * request, char *service,
   if ((sock = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
     {
       request->logging_function (gftp_logging_error, request,
-                                 _("Failed to create a socket: %s\n"),
+                                 _("Failed to create a IPv4 socket: %s\n"),
                                  g_strerror (errno));
       return (GFTP_ERETRYABLE);
     }
@@ -1849,16 +2076,14 @@ gftp_connect_server (gftp_request * request, char *service,
           close (sock);
           return (GFTP_EFATAL);
         }
-      else
-        {
-          port = serv_struct.s_port;
-          request->port = ntohs (serv_struct.s_port);
-        }
+
+      port = ntohs (serv_struct.s_port);
 
       if (!request->use_proxy)
-        request->port = ntohs (port);
+        request->port = port;
     }
-  remote_address.sin_port = port;
+
+  remote_address.sin_port = htons (port);
 
   if (request->hostp == NULL)
     {
@@ -1883,7 +2108,7 @@ gftp_connect_server (gftp_request * request, char *service,
               request->host.h_length);
       request->logging_function (gftp_logging_misc, request,
                                  _("Trying %s:%d\n"),
-                                 request->host.h_name, ntohs (port));
+                                 request->host.h_name, port);
 
       if (connect (sock, (struct sockaddr *) &remote_address,
                    sizeof (remote_address)) == -1)
@@ -1900,7 +2125,6 @@ gftp_connect_server (gftp_request * request, char *service,
       close (sock);
       return (GFTP_ERETRYABLE);
     }
-  port = ntohs (port);
 #endif /* HAVE_GETADDRINFO */
 
   if (fcntl (sock, F_SETFD, 1) == -1)
@@ -1952,12 +2176,15 @@ print_file_list (GList * list)
       tempfle = templist->data;
 #if defined (_LARGEFILE_SOURCE)
       printf ("%s:%s:%lld:%lld:%s:%s:%s\n", 
+              tempfle->file, tempfle->destfile, 
+              (long long) tempfle->size, (long long) tempfle->startsize, 
+              tempfle->user, tempfle->group, tempfle->attribs);
 #else
       printf ("%s:%s:%ld:%ld:%s:%s:%s\n", 
-#endif
               tempfle->file, tempfle->destfile, 
               tempfle->size, tempfle->startsize, 
               tempfle->user, tempfle->group, tempfle->attribs);
+#endif
       if (templist->next == NULL)
         break;
     }
@@ -1968,12 +2195,15 @@ print_file_list (GList * list)
       tempfle = templist->data;
 #if defined (_LARGEFILE_SOURCE)
       printf ("%s:%s:%lld:%lld:%s:%s:%s\n", 
+              tempfle->file, tempfle->destfile, 
+              (long long) tempfle->size, (long long) tempfle->startsize, 
+              tempfle->user, tempfle->group, tempfle->attribs);
 #else
       printf ("%s:%s:%ld:%ld:%s:%s:%s\n", 
-#endif
               tempfle->file, tempfle->destfile, 
               tempfle->size, tempfle->startsize, 
               tempfle->user, tempfle->group, tempfle->attribs);
+#endif
       if (templist == list)
         break;
     }
@@ -2109,7 +2339,7 @@ gftp_get_line (gftp_request * request, gftp_getline_buffer ** rbuf,
 ssize_t 
 gftp_fd_read (gftp_request * request, void *ptr, size_t size, int fd)
 {
-  long network_timeout;
+  intptr_t network_timeout;
   struct timeval tv;
   fd_set fset;
   ssize_t ret;
@@ -2179,7 +2409,7 @@ gftp_fd_read (gftp_request * request, void *ptr, size_t size, int fd)
 ssize_t 
 gftp_fd_write (gftp_request * request, const char *ptr, size_t size, int fd)
 {
-  long network_timeout;
+  intptr_t network_timeout;
   struct timeval tv;
   ssize_t w_ret;
   fd_set fset;
@@ -2310,10 +2540,17 @@ gftp_swap_socks (gftp_request * dest, gftp_request * source)
 
   dest->datafd = source->datafd;
   dest->cached = 0;
+#ifdef USE_SSL
+  dest->ssl = source->ssl;
+#endif
+
   if (!source->always_connected)
     {
       source->datafd = -1;
       source->cached = 1;
+#ifdef USE_SSL
+      source->ssl = NULL;
+#endif
     }
 
   if (dest->swap_socks)
@@ -2382,7 +2619,8 @@ gftp_calc_kbs (gftp_transfer * tdata, ssize_t num_read)
 int
 gftp_get_transfer_status (gftp_transfer * tdata, ssize_t num_read)
 {
-  int ret1, ret2, retries, sleep_time;
+  int ret1, ret2;
+  intptr_t retries, sleep_time;
   gftp_file * tempfle;
   struct timeval tv;
 
@@ -2413,33 +2651,38 @@ gftp_get_transfer_status (gftp_transfer * tdata, ssize_t num_read)
     {
       if (num_read == GFTP_EFATAL)
         return (GFTP_EFATAL);
-      else if (retries != 0 && 
-               tdata->current_file_retries >= retries)
+      else if (!tdata->conn_error_no_timeout)
         {
-          tdata->fromreq->logging_function (gftp_logging_error, tdata->fromreq,
-                   _("Error: Remote site %s disconnected. Max retries reached...giving up\n"),
-                   tdata->fromreq->hostname != NULL ? 
-                         tdata->fromreq->hostname : tdata->toreq->hostname);
-          return (GFTP_EFATAL);
-        }
-      else
-        {
-          tdata->fromreq->logging_function (gftp_logging_error, tdata->fromreq,
-                     _("Error: Remote site %s disconnected. Will reconnect in %d seconds\n"),
-                     tdata->fromreq->hostname != NULL ? 
-                           tdata->fromreq->hostname : tdata->toreq->hostname, 
-                     sleep_time);
+          if (retries != 0 && 
+              tdata->current_file_retries >= retries)
+            {
+              tdata->fromreq->logging_function (gftp_logging_error, tdata->fromreq,
+                       _("Error: Remote site %s disconnected. Max retries reached...giving up\n"),
+                       tdata->fromreq->hostname != NULL ? 
+                             tdata->fromreq->hostname : tdata->toreq->hostname);
+              return (GFTP_EFATAL);
+            }
+          else
+            {
+              tdata->fromreq->logging_function (gftp_logging_error, tdata->fromreq,
+                         _("Error: Remote site %s disconnected. Will reconnect in %d seconds\n"),
+                         tdata->fromreq->hostname != NULL ? 
+                             tdata->fromreq->hostname : tdata->toreq->hostname, 
+                         sleep_time);
+            }
         }
 
       while (retries == 0 || 
              tdata->current_file_retries <= retries)
         {
-          if (!tdata->skip_file)
+          if (!tdata->conn_error_no_timeout && !tdata->skip_file)
             {
               tv.tv_sec = sleep_time;
               tv.tv_usec = 0;
               select (0, NULL, NULL, NULL, &tv);
             }
+          else
+            tdata->conn_error_no_timeout = 0;
 
           if ((ret1 = gftp_connect (tdata->fromreq)) == 0 &&
               (ret2 = gftp_connect (tdata->toreq)) == 0)
@@ -2528,3 +2771,20 @@ gftp_fd_open (gftp_request * request, const char *pathname, int flags, mode_t mo
 
   return (fd);
 }
+
+
+void
+gftp_setup_startup_directory (gftp_request * request)
+{
+  char *startup_directory, *tempstr;
+
+  gftp_lookup_request_option (request, "startup_directory", &startup_directory);
+
+  if (*startup_directory != '\0' &&
+      (tempstr = expand_path (startup_directory)) != NULL)
+    {
+      gftp_set_directory (request, tempstr);
+      g_free (tempstr);
+    }
+}
+
