@@ -18,7 +18,9 @@
 /*****************************************************************************/
 
 #include "gftp.h"
-static const char cvsid[] = "$Id: rfc959.c,v 1.37 2003/08/12 00:38:05 masneyb Exp $";
+#include "ftpcommon.h"
+
+static const char cvsid[] = "$Id: rfc959.c,v 1.66 2004/03/26 21:03:50 masneyb Exp $";
 
 static gftp_textcomboedt_data gftp_proxy_type[] = {
   {N_("none"), "", 0},
@@ -84,24 +86,14 @@ static gftp_config_vars config_vars[] =
 };
 
          
-typedef struct rfc959_params_tag
-{
-  gftp_getline_buffer * datafd_rbuf,
-                      * dataconn_rbuf;
-  int data_connection;
-  unsigned int is_ascii_transfer : 1;
-} rfc959_parms;
-
-
 static int
-rfc959_read_response (gftp_request * request)
+rfc959_read_response (gftp_request * request, int disconnect_on_42x)
 {
   char tempstr[255], code[4];
   rfc959_parms * parms;
   ssize_t num_read;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   *code = '\0';
@@ -136,20 +128,24 @@ rfc959_read_response (gftp_request * request)
   request->last_ftp_response = g_strdup (tempstr);
 
   if (request->last_ftp_response[0] == '4' &&
-      request->last_ftp_response[1] == '2')
-    gftp_disconnect (request);
+      request->last_ftp_response[1] == '2' &&
+      disconnect_on_42x)
+    {
+      gftp_disconnect (request);
+      return (GFTP_ETIMEDOUT);
+    }
 
   return (*request->last_ftp_response);
 }
 
 
-static int
-rfc959_send_command (gftp_request * request, const char *command)
+int
+rfc959_send_command (gftp_request * request, const char *command, 
+                     int read_response)
 {
   int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (command != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
@@ -169,24 +165,26 @@ rfc959_send_command (gftp_request * request, const char *command)
                                  command);
     }
 
-  if ((ret = gftp_fd_write (request, command, strlen (command), 
-                         request->datafd)) < 0)
+  if ((ret = request->write_function (request, command, strlen (command), 
+                                      request->datafd)) < 0)
     return (ret);
 
-  return (rfc959_read_response (request));
+  if (read_response)
+    return (rfc959_read_response (request, 1));
+  else
+    return (0);
 }
 
 
 static char *
 parse_ftp_proxy_string (gftp_request * request)
 {
-  char *startpos, *endpos, *newstr, *newval, tempport[6], *proxy_config,
+  char *startpos, *endpos, *newstr, *newval, tempport[6], *proxy_config, *utf8,
        savechar;
   size_t len;
-  int tmp;
+  intptr_t tmp;
 
   g_return_val_if_fail (request != NULL, NULL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, NULL);
 
   gftp_lookup_request_option (request, "proxy_config", &proxy_config);
 
@@ -210,7 +208,7 @@ parse_ftp_proxy_string (gftp_request * request)
 	      break;
 	    case 'o':
               gftp_lookup_request_option (request, "ftp_proxy_port", &tmp);
-              g_snprintf (tempport, sizeof (tempport), "%d", tmp);
+              g_snprintf (tempport, sizeof (tempport), "%d", (int)tmp);
 	      newval = tempport;
 	      break;
 	    case 'a':
@@ -277,10 +275,22 @@ parse_ftp_proxy_string (gftp_request * request)
         }
       else
         {
-          len += strlen (newval);
+          utf8 = gftp_string_from_utf8 (request, newval);
+          if (utf8 != NULL)
+            len += strlen (utf8);
+          else
+            len += strlen (newval);
+
           newstr = g_realloc (newstr, sizeof (char) * (len + 1));
           strcat (newstr, startpos);
-          strcat (newstr, newval);
+
+          if (utf8 != NULL)
+            {
+              strcat (newstr, utf8);
+              g_free (utf8);
+            }
+          else
+            strcat (newstr, newval);
         }
    
       *endpos = savechar;
@@ -298,14 +308,14 @@ rfc959_getcwd (gftp_request * request)
   char *pos, *dir;
   int ret;
 
-  ret = rfc959_send_command (request, "PWD\r\n");
+  ret = rfc959_send_command (request, "PWD\r\n", 1);
   if (ret < 0)
     return (ret);
   else if (ret != '2')
     {
       request->logging_function (gftp_logging_error, request,
-				 _("Received invalid response to PWD command: '%s'\n"),
-                                 request->last_ftp_response);
+                                 _("Invalid response '%c' received from server.\n"),
+                                 ret);
       gftp_disconnect (request);
       return (GFTP_ERETRYABLE);
     }
@@ -313,8 +323,8 @@ rfc959_getcwd (gftp_request * request)
   if ((pos = strchr (request->last_ftp_response, '"')) == NULL)
     {
       request->logging_function (gftp_logging_error, request,
-				 _("Received invalid response to PWD command: '%s'\n"),
-                                 request->last_ftp_response);
+                                 _("Invalid response '%c' received from server.\n"),
+                                 ret);
       gftp_disconnect (request);
       return (GFTP_EFATAL);
     }
@@ -324,8 +334,8 @@ rfc959_getcwd (gftp_request * request)
   if ((pos = strchr (dir, '"')) == NULL)
     {
       request->logging_function (gftp_logging_error, request,
-				 _("Received invalid response to PWD command: '%s'\n"),
-                                 request->last_ftp_response);
+                                 _("Invalid response '%c' received from server.\n"),
+                                 ret);
       gftp_disconnect (request);
       return (GFTP_EFATAL);
     }
@@ -347,19 +357,20 @@ rfc959_chdir (gftp_request * request, const char *directory)
   int r;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (directory != NULL, GFTP_EFATAL);
 
   if (strcmp (directory, "..") == 0)
-    ret = rfc959_send_command (request, "CDUP\r\n");
+    ret = rfc959_send_command (request, "CDUP\r\n", 1);
   else
     {
       tempstr = g_strconcat ("CWD ", directory, "\r\n", NULL);
-      ret = rfc959_send_command (request, tempstr);
+      ret = rfc959_send_command (request, tempstr, 1);
       g_free (tempstr);
     }
 
-  if (ret != '2')
+  if (ret < 0)
+    return (ret);
+  else if (ret != '2')
     return (GFTP_ERETRYABLE);
 
   if (directory != request->directory)
@@ -379,10 +390,9 @@ rfc959_syst (gftp_request * request)
   int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
-  ret = rfc959_send_command (request, "SYST\r\n");
+  ret = rfc959_send_command (request, "SYST\r\n", 1);
 
   if (ret < 0)
     return (ret);
@@ -402,6 +412,11 @@ rfc959_syst (gftp_request * request)
     request->server_type = GFTP_DIRTYPE_UNIX;
   else if (strcmp (stpos, "VMS") == 0)
     request->server_type = GFTP_DIRTYPE_VMS;
+  else if (strcmp (stpos, "MVS") == 0 ||
+           strcmp (stpos, "OS/MVS") == 0)
+    request->server_type = GFTP_DIRTYPE_MVS;
+  else if (strcmp (stpos, "NETWARE") == 0)
+    request->server_type = GFTP_DIRTYPE_NOVELL;
   else if (strcmp (stpos, "CRAY") == 0)
     request->server_type = GFTP_DIRTYPE_CRAY;
   else
@@ -411,15 +426,15 @@ rfc959_syst (gftp_request * request)
 }
 
 
-static int
+int
 rfc959_connect (gftp_request * request)
 {
-  char tempchar, *startpos, *endpos, *tempstr, *email, *proxy_hostname;
-  int ret, resp, ascii_transfers, proxy_port;
+  char tempchar, *startpos, *endpos, *tempstr, *email, *proxy_hostname, *utf8;
+  int ret, resp;
+  intptr_t ascii_transfers, proxy_port;
   rfc959_parms * parms;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (request->hostname != NULL, GFTP_EFATAL);
 
   if (request->datafd > 0)
@@ -427,28 +442,38 @@ rfc959_connect (gftp_request * request)
 
   parms = request->protocol_data;
 
-  gftp_lookup_request_option (request, "email", &email);
   gftp_lookup_request_option (request, "ftp_proxy_host", &proxy_hostname);
   gftp_lookup_request_option (request, "ftp_proxy_port", &proxy_port);
 
   if (request->username == NULL || *request->username == '\0')
+    gftp_set_username (request, "anonymous");
+
+  if (strcasecmp (request->username, "anonymous") == 0 &&
+      (request->password == NULL || *request->password == '\0'))
     {
-      gftp_set_username (request, "anonymous");
+      gftp_lookup_request_option (request, "email", &email);
       gftp_set_password (request, email);
     }
-  else if (strcasecmp (request->username, "anonymous") == 0)
-    gftp_set_password (request, email);
    
   if ((ret = gftp_connect_server (request, "ftp", proxy_hostname, proxy_port)) < 0)
     return (ret);
 
   /* Get the banner */
-  if ((ret = rfc959_read_response (request)) != '2')
+  if ((ret = rfc959_read_response (request, 1)) != '2')
     {
       gftp_disconnect (request);
       return (ret);
     }
 
+  if (parms->auth_tls_start != NULL)
+    {
+      if ((ret = parms->auth_tls_start (request)) < 0)
+        {
+          gftp_disconnect (request);
+          return (ret);
+        }
+    }
+
   /* Login the proxy server if available */
   if (request->use_proxy)
     {
@@ -461,7 +486,7 @@ rfc959_connect (gftp_request * request)
 	      tempchar = *(endpos + 1);
 	      if (*endpos != '\0')
 		*(endpos + 1) = '\0';
-	      if ((resp = rfc959_send_command (request, startpos)) < 0)
+	      if ((resp = rfc959_send_command (request, startpos, 1)) < 0)
                 return (resp);
 	      if (*endpos != '\0')
 		*(endpos + 1) = tempchar;
@@ -476,32 +501,46 @@ rfc959_connect (gftp_request * request)
   else
     {
       tempstr = g_strconcat ("USER ", request->username, "\r\n", NULL);
-      resp = rfc959_send_command (request, tempstr);
+      resp = rfc959_send_command (request, tempstr, 1);
       g_free (tempstr);
       if (resp < 0)
-        return (GFTP_ERETRYABLE);
+        return (resp);
+
       if (resp == '3')
 	{
-	  tempstr = g_strconcat ("PASS ", request->password, "\r\n", NULL);
-	  resp = rfc959_send_command (request, tempstr);
+          utf8 = gftp_string_from_utf8 (request, request->password);
+          if (utf8 != NULL)
+            {
+              tempstr = g_strconcat ("PASS ", utf8, "\r\n", NULL);
+              g_free (utf8);
+            }
+          else
+            tempstr = g_strconcat ("PASS ", request->password, "\r\n", NULL);
+
+	  resp = rfc959_send_command (request, tempstr, 1);
 	  g_free (tempstr);
           if (resp < 0)
-            return (GFTP_ERETRYABLE);
+            return (resp);
         }
+
       if (resp == '3' && request->account)
 	{
 	  tempstr = g_strconcat ("ACCT ", request->account, "\r\n", NULL);
-	  resp = rfc959_send_command (request, tempstr);
+	  resp = rfc959_send_command (request, tempstr, 1);
 	  g_free (tempstr);
           if (resp < 0)
-            return (GFTP_ERETRYABLE);
+            return (resp);
 	}
     }
 
   if (resp != '2')
     {
       gftp_disconnect (request);
-      return (GFTP_EFATAL);
+
+      if (resp == '5')
+        return (GFTP_EFATAL);
+      else
+        return (GFTP_ERETRYABLE);
     }
 
   if ((ret = rfc959_syst (request)) < 0 && request->datafd < 0)
@@ -519,7 +558,7 @@ rfc959_connect (gftp_request * request)
       parms->is_ascii_transfer = 0;
     }
 
-  if ((ret = rfc959_send_command (request, tempstr)) < 0)
+  if ((ret = rfc959_send_command (request, tempstr, 1)) < 0)
     return (ret);
 
   ret = -1;
@@ -549,7 +588,6 @@ rfc959_disconnect (gftp_request * request)
   rfc959_parms * parms;
 
   g_return_if_fail (request != NULL);
-  g_return_if_fail (request->protonum == GFTP_FTP_NUM);
 
   parms = request->protocol_data;
 
@@ -575,7 +613,8 @@ rfc959_ipv4_data_connection_new (gftp_request * request)
 {
   char *pos, *pos1, resp, *command;
   struct sockaddr_in data_addr;
-  int i, passive_transfer;
+  int i;
+  intptr_t passive_transfer;
   rfc959_parms * parms;
   socklen_t data_addr_len;
   unsigned int temp[6];
@@ -586,7 +625,7 @@ rfc959_ipv4_data_connection_new (gftp_request * request)
   if ((parms->data_connection = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
     {
       request->logging_function (gftp_logging_error, request,
-				 _("Failed to create a socket: %s\n"),
+				 _("Failed to create a IPv4 socket: %s\n"),
 				 g_strerror (errno));
       gftp_disconnect (request);
       return (GFTP_ERETRYABLE);
@@ -608,12 +647,13 @@ rfc959_ipv4_data_connection_new (gftp_request * request)
   gftp_lookup_request_option (request, "passive_transfer", &passive_transfer);
   if (passive_transfer)
     {
-      if ((resp = rfc959_send_command (request, "PASV\r\n")) != '2')
+      resp = rfc959_send_command (request, "PASV\r\n", 1);
+      if (resp < 0)
+        return (resp);
+      else if (resp != '2')
 	{
-          if (request->datafd < 0)
-            return (resp);
-
-          gftp_set_request_option (request, "passive_transfer", GINT_TO_POINTER(0));
+          gftp_set_request_option (request, "passive_transfer",
+                                   GINT_TO_POINTER(0));
 	  return (rfc959_ipv4_data_connection_new (request));
 	}
 
@@ -704,10 +744,16 @@ rfc959_ipv4_data_connection_new (gftp_request * request)
 				 pos[0] & 0xff, pos[1] & 0xff, pos[2] & 0xff,
 				 pos[3] & 0xff, pos1[0] & 0xff,
 				 pos1[1] & 0xff);
-      resp = rfc959_send_command (request, command);
+      resp = rfc959_send_command (request, command, 1);
       g_free (command);
-      if (resp != '2')
+
+      if (resp < 0)
+        return (resp);
+      else if (resp != '2')
 	{
+          request->logging_function (gftp_logging_error, request,
+                                     _("Invalid response '%c' received from server.\n"),
+                                     resp);
           gftp_disconnect (request);
 	  return (GFTP_ERETRYABLE);
 	}
@@ -733,7 +779,7 @@ rfc959_ipv6_data_connection_new (gftp_request * request)
   if ((parms->data_connection = socket (AF_INET6, SOCK_STREAM, IPPROTO_TCP)) < 0)
     {
       request->logging_function (gftp_logging_error, request,
-				 _("Failed to create a socket: %s\n"),
+				 _("Failed to create a IPv6 socket: %s\n"),
 				 g_strerror (errno));
       gftp_disconnect (request);
       return (GFTP_ERETRYABLE);
@@ -764,11 +810,11 @@ rfc959_ipv6_data_connection_new (gftp_request * request)
   gftp_lookup_request_option (request, "passive_transfer", &passive_transfer);
   if (passive_transfer)
     {
-      if ((resp = rfc959_send_command (request, "EPSV\r\n")) != '2')
+      resp = rfc959_send_command (request, "EPSV\r\n", 1);
+      if (resp < 0)
+        return (resp);
+      else if (resp != '2')
 	{
-          if (request->datafd < 0)
-            return (resp);
-
           gftp_set_request_option (request, "passive_transfer", 
                                    GINT_TO_POINTER(0));
 	  return (rfc959_ipv6_data_connection_new (request));
@@ -788,7 +834,7 @@ rfc959_ipv6_data_connection_new (gftp_request * request)
           return (GFTP_EFATAL);
         }
 
-      if (sscanf (pos, "|||%d|", &port) != 1)
+      if (sscanf (pos, "|||%u|", &port) != 1)
         {
           request->logging_function (gftp_logging_error, request,
                       _("Invalid EPSV response '%s'\n"),
@@ -857,9 +903,12 @@ rfc959_ipv6_data_connection_new (gftp_request * request)
       command = g_strdup_printf ("EPRT |2|%s|%d|\n", buf,
                                  ntohs (data_addr.sin6_port));
 
-      resp = rfc959_send_command (request, command);
+      resp = rfc959_send_command (request, command, 1);
       g_free (command);
-      if (resp != '2')
+
+      if (resp < 0)
+        return (resp);
+      else if (resp != '2')
 	{
           gftp_disconnect (request);
 	  return (GFTP_ERETRYABLE);
@@ -876,7 +925,6 @@ static int
 rfc959_data_connection_new (gftp_request * request)
 {
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
 #ifdef HAVE_IPV6
@@ -893,19 +941,19 @@ rfc959_data_connection_new (gftp_request * request)
 static int
 rfc959_accept_active_connection (gftp_request * request)
 {
-  int infd, ret, passive_transfer;
+  int infd, ret;
+  intptr_t passive_transfer;
   rfc959_parms * parms;
 #ifdef HAVE_IPV6
-  struct sockaddr_in cli_addr;
-#else
   struct sockaddr_in6 cli_addr;
+#else
+  struct sockaddr_in cli_addr;
 #endif
   socklen_t cli_addr_len;
 
   parms = request->protocol_data;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (parms->data_connection > 0, GFTP_EFATAL);
 
   gftp_lookup_request_option (request, "passive_transfer", &passive_transfer);
@@ -941,7 +989,8 @@ rfc959_is_ascii_transfer (gftp_request * request, const char *filename)
 {
   gftp_config_list_vars * tmplistvar;
   gftp_file_extensions * tempext;
-  int stlen, ascii_transfers;
+  int stlen;
+  intptr_t ascii_transfers;
   GList * templist;
   
   gftp_lookup_global_option ("ext", &tmplistvar);
@@ -967,15 +1016,14 @@ rfc959_is_ascii_transfer (gftp_request * request, const char *filename)
 }
 
 
-static void
+static int
 rfc959_set_data_type (gftp_request * request, const char *filename)
 {
   rfc959_parms * parms;
-  int new_ascii;
+  int new_ascii, ret;
   char *tempstr;
 
-  g_return_if_fail (request != NULL);
-  g_return_if_fail (request->protonum == GFTP_FTP_NUM);
+  g_return_val_if_fail (request != NULL, GFTP_EFATAL);
 
   parms = request->protocol_data;
   new_ascii = rfc959_is_ascii_transfer (request, filename);
@@ -993,10 +1041,11 @@ rfc959_set_data_type (gftp_request * request, const char *filename)
           parms->is_ascii_transfer = 0;
         }
 
-      rfc959_send_command (request, tempstr);
+      if ((ret = rfc959_send_command (request, tempstr, 1)) < 0)
+        return (ret);
     }
 
-  return;
+  return (0);
 }
 
 
@@ -1005,11 +1054,11 @@ rfc959_get_file (gftp_request * request, const char *filename, int fd,
                  off_t startsize)
 {
   char *command, *tempstr, resp;
-  int ret, passive_transfer;
+  int ret;
+  intptr_t passive_transfer;
   rfc959_parms * parms;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (filename != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
@@ -1017,7 +1066,8 @@ rfc959_get_file (gftp_request * request, const char *filename, int fd,
   if (fd > 0)
     parms->data_connection = fd;
 
-  rfc959_set_data_type (request, filename);
+  if ((ret = rfc959_set_data_type (request, filename)) < 0)
+    return (ret);
 
   if (parms->data_connection < 0 && 
       (ret = rfc959_data_connection_new (request)) < 0)
@@ -1029,14 +1079,16 @@ rfc959_get_file (gftp_request * request, const char *filename, int fd,
   if (startsize > 0)
     {
 #if defined (_LARGEFILE_SOURCE)
-      command = g_strdup_printf ("REST %lld\r\n", startsize); 
+      command = g_strdup_printf ("REST %lld\r\n", (long long) startsize); 
 #else
       command = g_strdup_printf ("REST %ld\r\n", startsize); 
 #endif
-      resp = rfc959_send_command (request, command);
+      resp = rfc959_send_command (request, command, 1);
       g_free (command);
 
-      if (resp != '3')
+      if (resp < 0)
+        return (resp);
+      else if (resp != '3')
         {
           close (parms->data_connection);
           parms->data_connection = -1;
@@ -1045,10 +1097,12 @@ rfc959_get_file (gftp_request * request, const char *filename, int fd,
     }
 
   tempstr = g_strconcat ("RETR ", filename, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
-  if (ret != '1')
+  if (ret < 0)
+    return (ret);
+  else if (ret != '1')
     {
       close (parms->data_connection);
       parms->data_connection = -1;
@@ -1078,11 +1132,11 @@ rfc959_put_file (gftp_request * request, const char *filename, int fd,
                  off_t startsize, off_t totalsize)
 {
   char *command, *tempstr, resp;
-  int ret, passive_transfer;
+  intptr_t passive_transfer;
   rfc959_parms * parms;
+  int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (filename != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
@@ -1090,7 +1144,8 @@ rfc959_put_file (gftp_request * request, const char *filename, int fd,
   if (fd > 0)
     fd = parms->data_connection;
 
-  rfc959_set_data_type (request, filename);
+  if ((ret = rfc959_set_data_type (request, filename)) < 0)
+    return (ret);
 
   if (parms->data_connection < 0 && 
       (ret = rfc959_data_connection_new (request)) < 0)
@@ -1102,13 +1157,15 @@ rfc959_put_file (gftp_request * request, const char *filename, int fd,
   if (startsize > 0)
     {
 #if defined (_LARGEFILE_SOURCE)
-      command = g_strdup_printf ("REST %lld\r\n", startsize); 
+      command = g_strdup_printf ("REST %lld\r\n", (long long) startsize); 
 #else
       command = g_strdup_printf ("REST %ld\r\n", startsize); 
 #endif
-      resp = rfc959_send_command (request, command);
+      resp = rfc959_send_command (request, command, 1);
       g_free (command);
-      if (resp != '3')
+      if (resp < 0)
+        return (resp);
+      else if (resp != '3')
         {
           close (parms->data_connection);
           parms->data_connection = -1;
@@ -1117,9 +1174,11 @@ rfc959_put_file (gftp_request * request, const char *filename, int fd,
     }
 
   tempstr = g_strconcat ("STOR ", filename, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
-  if (ret != '1')
+  if (ret < 0)
+    return (ret);
+  else if (ret != '1')
     {
       close (parms->data_connection);
       parms->data_connection = -1;
@@ -1141,6 +1200,7 @@ rfc959_transfer_file (gftp_request *fromreq, const char *fromfile,
                       const char *tofile, off_t tosize)
 {
   char *tempstr, *pos, *endpos;
+  rfc959_parms * parms;
   int ret;
 
   g_return_val_if_fail (fromreq != NULL, GFTP_EFATAL);
@@ -1150,11 +1210,10 @@ rfc959_transfer_file (gftp_request *fromreq, const char *fromfile,
   g_return_val_if_fail (fromreq->datafd > 0, GFTP_EFATAL);
   g_return_val_if_fail (toreq->datafd > 0, GFTP_EFATAL);
 
-  gftp_set_request_option (fromreq, "passive_transfer", GINT_TO_POINTER(1));
-  gftp_set_request_option (toreq, "passive_transfer", GINT_TO_POINTER(0));
-
-  if ((ret = rfc959_send_command (fromreq, "PASV\r\n")) != '2')
+  if ((ret = rfc959_send_command (fromreq, "PASV\r\n", 1)) < 0)
     return (ret);
+  else if (ret != '2')
+    return (GFTP_ERETRYABLE);
 
   pos = fromreq->last_ftp_response + 4;
   while (!isdigit ((int) *pos) && *pos != '\0') 
@@ -1169,37 +1228,40 @@ rfc959_transfer_file (gftp_request *fromreq, const char *fromfile,
     *endpos = '\0';
 
   tempstr = g_strconcat ("PORT ", pos, "\r\n", NULL);
-  if ((ret = rfc959_send_command (toreq, tempstr)) != '2')
-     {
-       g_free (tempstr);
-       return (ret);
-     }
+  ret = rfc959_send_command (toreq, tempstr, 1);
   g_free (tempstr);
 
+  if (ret < 0)
+    return (ret);
+  else if (ret != '2')
+    return (GFTP_ERETRYABLE);
+
   tempstr = g_strconcat ("RETR ", fromfile, "\r\n", NULL);
-  if ((ret = gftp_fd_write (fromreq, tempstr, strlen (tempstr), 
-                         fromreq->datafd)) < 0)
-    {
-      g_free (tempstr);
-      return (ret);
-    }
+  ret = rfc959_send_command (fromreq, tempstr, 0);
   g_free (tempstr);
 
+  if (ret < 0)
+    return (ret);
+
   tempstr = g_strconcat ("STOR ", tofile, "\r\n", NULL);
-  if ((ret = gftp_fd_write (toreq, tempstr, strlen (tempstr), 
-                            toreq->datafd)) < 0)
-    {
-      g_free (tempstr);
-      return (ret);
-    }
+  ret = rfc959_send_command (toreq, tempstr, 0);
   g_free (tempstr);
 
-  if ((ret = rfc959_read_response (fromreq)) < 0)
+  if (ret < 0)
     return (ret);
 
-  if ((ret = rfc959_read_response (toreq)) < 0)
+  if ((ret = rfc959_read_response (fromreq, 1)) < 0)
     return (ret);
 
+  if ((ret = rfc959_read_response (toreq, 1)) < 0)
+    return (ret);
+
+  parms = fromreq->protocol_data;
+  parms->is_fxp_transfer = 1;
+
+  parms = toreq->protocol_data;
+  parms->is_fxp_transfer = 1;
+
   return (0);
 }
 
@@ -1211,17 +1273,18 @@ rfc959_end_transfer (gftp_request * request)
   int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   parms = request->protocol_data;
+  parms->is_fxp_transfer = 0;
+
   if (parms->data_connection > 0)
     {
       close (parms->data_connection);
       parms->data_connection = -1;
     }
 
-  ret = rfc959_read_response (request);
+  ret = rfc959_read_response (request, 1);
 
   if (ret < 0)
     return (ret);
@@ -1239,28 +1302,25 @@ rfc959_abort_transfer (gftp_request * request)
   int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   parms = request->protocol_data;
 
+  if ((ret = rfc959_send_command (request, "ABOR\r\n", 0)) < 0)
+    return (ret);
+
   if (parms->data_connection > 0)
     {
       close (parms->data_connection);
       parms->data_connection = -1;
     }
 
-  /* We need to read two lines of output. The first one is acknowleging
-     the transfer and the second line acknowleges the ABOR command */
-  if ((ret = rfc959_send_command (request, "ABOR\r\n")) < 0)
-    return (ret);
-
   if (request->datafd > 0)
     {
-      if ((ret = rfc959_read_response (request)) < 0)
+      if ((ret = rfc959_read_response (request, 0)) < 0)
         gftp_disconnect (request);
     }
-  
+
   return (0);
 }
 
@@ -1268,11 +1328,11 @@ rfc959_abort_transfer (gftp_request * request)
 static int
 rfc959_list_files (gftp_request * request)
 {
-  int ret, show_hidden_files, resolve_symlinks, passive_transfer;
+  int ret;
+  intptr_t show_hidden_files, resolve_symlinks, passive_transfer;
   char *tempstr, parms[3];
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   if ((ret = rfc959_data_connection_new (request)) < 0)
@@ -1288,11 +1348,18 @@ rfc959_list_files (gftp_request * request)
   tempstr = g_strconcat ("LIST", *parms != '\0' ? " -" : "", parms, "\r\n", 
                          NULL); 
 
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
-  if (ret != '1')
-    return (GFTP_ERETRYABLE);
+  if (ret < 0)
+    return (ret);
+  else if (ret != '1')
+    {
+      request->logging_function (gftp_logging_error, request,
+                                 _("Invalid response '%c' received from server.\n"),
+                                 ret);
+      return (GFTP_ERETRYABLE);
+    }
 
   ret = 0;
   if (!passive_transfer)
@@ -1305,47 +1372,52 @@ rfc959_list_files (gftp_request * request)
 static ssize_t
 rfc959_get_next_file_chunk (gftp_request * request, char *buf, size_t size)
 {
-  int i, j, ascii_transfers;
+  ssize_t num_read, ret;
   rfc959_parms * parms;
-  ssize_t num_read;
+  int i, j;
 
   parms = request->protocol_data;
-  num_read = gftp_fd_read (request, buf, size, parms->data_connection);
+  if (parms->is_fxp_transfer)
+    return (GFTP_ENOTRANS);
+
+  num_read = parms->data_conn_read (request, buf, size, parms->data_connection);
   if (num_read < 0)
     return (num_read);
 
-  gftp_lookup_request_option (request, "ascii_transfers", &ascii_transfers);
-  if (ascii_transfers)
+  ret = num_read;
+  if (parms->is_ascii_transfer)
     {
       for (i = 0, j = 0; i < num_read; i++)
         {
           if (buf[i] != '\r')
             buf[j++] = buf[i];
           else
-            num_read--;
+            ret--;
         }
     }
 
-  return (num_read);
+  return (ret);
 }
 
 
 static ssize_t
 rfc959_put_next_file_chunk (gftp_request * request, char *buf, size_t size)
 {
-  int i, j, ascii_transfers;
   rfc959_parms * parms;
   ssize_t num_wrote;
   char *tempstr;
   size_t rsize;
+  int i, j;
 
   if (size == 0)
     return (0);
 
   parms = request->protocol_data;
 
-  gftp_lookup_request_option (request, "ascii_transfers", &ascii_transfers);
-  if (ascii_transfers)
+  if (parms->is_fxp_transfer)
+    return (GFTP_ENOTRANS);
+
+  if (parms->is_ascii_transfer)
     {
       rsize = 0;
       for (i = 0; i < size; i++)
@@ -1375,7 +1447,8 @@ rfc959_put_next_file_chunk (gftp_request * request, char *buf, size_t size)
       tempstr = buf;
     }
 
-  num_wrote = gftp_fd_write (request, tempstr, rsize, parms->data_connection);
+  num_wrote = parms->data_conn_write (request, tempstr, rsize,
+                                      parms->data_connection);
 
   if (tempstr != buf)
     g_free (tempstr);
@@ -1387,12 +1460,13 @@ rfc959_put_next_file_chunk (gftp_request * request, char *buf, size_t size)
 int
 rfc959_get_next_file (gftp_request * request, gftp_file * fle, int fd)
 {
+  ssize_t (*oldread_func) (gftp_request * request, void *ptr, size_t size,
+                           int fd);
   rfc959_parms * parms;
   char tempstr[1024];
   ssize_t len;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (fle != NULL, GFTP_EFATAL);
   g_return_val_if_fail (fd > 0, GFTP_EFATAL);
 
@@ -1409,8 +1483,13 @@ rfc959_get_next_file (gftp_request * request, gftp_file * fle, int fd)
 
   do
     {
-      if ((len = gftp_get_line (request, &parms->dataconn_rbuf,
-                                tempstr, sizeof (tempstr), fd)) <= 0)
+      oldread_func = request->read_function;
+      request->read_function = parms->data_conn_read;
+      len = gftp_get_line (request, &parms->dataconn_rbuf, tempstr,
+                           sizeof (tempstr), fd);
+      request->read_function = oldread_func;
+
+      if (len <= 0)
 	{
           gftp_file_destroy (fle);
 	  return ((int) len);
@@ -1448,12 +1527,11 @@ rfc959_get_file_size (gftp_request * request, const char *filename)
   int ret;
 
   g_return_val_if_fail (request != NULL, 0);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (filename != NULL, 0);
   g_return_val_if_fail (request->datafd > 0, 0);
 
   tempstr = g_strconcat ("SIZE ", filename, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
   if (ret < 0)
     return (ret);
@@ -1470,12 +1548,11 @@ rfc959_rmdir (gftp_request * request, const char *directory)
   char *tempstr, ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (directory != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   tempstr = g_strconcat ("RMD ", directory, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
   if (ret < 0)
@@ -1493,12 +1570,11 @@ rfc959_rmfile (gftp_request * request, const char *file)
   char *tempstr, ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (file != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   tempstr = g_strconcat ("DELE ", file, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
   if (ret < 0)
@@ -1516,12 +1592,11 @@ rfc959_mkdir (gftp_request * request, const char *directory)
   char *tempstr, ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (directory != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   tempstr = g_strconcat ("MKD ", directory, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
   if (ret < 0)
@@ -1540,22 +1615,21 @@ rfc959_rename (gftp_request * request, const char *oldname,
   char *tempstr, ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (oldname != NULL, GFTP_EFATAL);
   g_return_val_if_fail (newname != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   tempstr = g_strconcat ("RNFR ", oldname, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
   if (ret < 0)
     return (ret);
-  else if (ret != '2')
+  else if (ret != '3')
     return (GFTP_ERETRYABLE);
 
   tempstr = g_strconcat ("RNTO ", newname, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
   if (ret < 0)
@@ -1573,13 +1647,12 @@ rfc959_chmod (gftp_request * request, const char *file, int mode)
   char *tempstr, ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (file != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   tempstr = g_malloc (strlen (file) + (mode / 10) + 16);
   sprintf (tempstr, "SITE CHMOD %d %s\r\n", mode, file);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
   if (ret < 0)
@@ -1597,12 +1670,11 @@ rfc959_site (gftp_request * request, const char *command)
   char *tempstr, ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
-  g_return_val_if_fail (request->protonum == GFTP_FTP_NUM, GFTP_EFATAL);
   g_return_val_if_fail (command != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->datafd > 0, GFTP_EFATAL);
 
   tempstr = g_strconcat ("SITE ", command, "\r\n", NULL);
-  ret = rfc959_send_command (request, tempstr);
+  ret = rfc959_send_command (request, tempstr, 1);
   g_free (tempstr);
 
   if (ret < 0)
@@ -1655,6 +1727,27 @@ rfc959_request_destroy (gftp_request * request)
 }
 
 
+void
+rfc959_copy_param_options (gftp_request * dest_request,
+                           gftp_request * src_request)
+{
+  rfc959_parms * dparms, * sparms;
+
+  dparms = dest_request->protocol_data;
+  sparms = src_request->protocol_data;
+
+  dparms->data_connection = -1;
+  dparms->is_ascii_transfer = sparms->is_ascii_transfer;
+  dparms->is_fxp_transfer = sparms->is_fxp_transfer;
+  dparms->auth_tls_start = sparms->auth_tls_start;
+  dparms->data_conn_read = sparms->data_conn_read;
+  dparms->data_conn_write = sparms->data_conn_write;
+
+  dest_request->read_function = src_request->read_function;
+  dest_request->write_function = src_request->write_function;
+}
+
+
 int
 rfc959_init (gftp_request * request)
 {
@@ -1684,6 +1777,7 @@ rfc959_init (gftp_request * request)
 
   request->protonum = GFTP_FTP_NUM;
   request->init = rfc959_init;
+  request->copy_param_options = rfc959_copy_param_options;
   request->destroy = rfc959_request_destroy; 
   request->read_function = gftp_fd_read;
   request->write_function = gftp_fd_write;
@@ -1715,12 +1809,14 @@ rfc959_init (gftp_request * request)
   request->need_hostport = 1;
   request->need_userpass = 1;
   request->use_cache = 1;
-  request->use_threads = 1;
   request->always_connected = 0;
 
   request->protocol_data = g_malloc0 (sizeof (rfc959_parms));
   parms = request->protocol_data;
   parms->data_connection = -1; 
+  parms->auth_tls_start = NULL;
+  parms->data_conn_read = gftp_fd_read;
+  parms->data_conn_write = gftp_fd_write;
 
   return (gftp_set_config_options (request));
 }
