@@ -18,7 +18,7 @@
 /*****************************************************************************/
 
 #include "gftp.h"
-static const char cvsid[] = "$Id: sshv2.c,v 1.30 2003/08/05 01:40:48 masneyb Exp $";
+static const char cvsid[] = "$Id: sshv2.c,v 1.38 2004/02/15 17:53:34 masneyb Exp $";
 
 #define SSH_MAX_HANDLE_SIZE		256
 #define SSH_MAX_STRING_SIZE		34000
@@ -46,7 +46,7 @@ static gftp_config_vars config_vars[] =
    gftp_option_type_checkbox, GINT_TO_POINTER(0), NULL, 
    GFTP_CVARS_FLAGS_SHOW_BOOKMARK,
    N_("Use the ssh-askpass utility to supply the remote password"), GFTP_PORT_GTK,
-   NULL},
+        NULL},
   {"sshv2_use_sftp_subsys", N_("Use SSH2 SFTP subsys"), 
    gftp_option_type_checkbox, GINT_TO_POINTER(0), NULL, 
    GFTP_CVARS_FLAGS_SHOW_BOOKMARK,
@@ -79,20 +79,22 @@ typedef struct sshv2_message_tag
 
 typedef struct sshv2_params_tag
 {
-  char handle[SSH_MAX_HANDLE_SIZE + 4]; /* We'll encode the ID in here too */
-  int handle_len,
-      dont_log_status : 1;              /* For uploading files */
+  char handle[SSH_MAX_HANDLE_SIZE + 4], /* We'll encode the ID in here too */
+       *read_buffer;
+
+  gint32 handle_len,
+         id,
+         count;
   sshv2_message message;
 
-  gint32 id,				
-         count;
-  unsigned int initialized : 1;
+  unsigned int initialized : 1,
+               dont_log_status : 1;              /* For uploading files */
+
 #ifdef G_HAVE_GINT64
   gint64 offset;
 #else
   gint32 offset;
 #endif
-  char *read_buffer;
 } sshv2_params;
 
 
@@ -562,7 +564,7 @@ sshv2_log_command (gftp_request * request, gftp_logging_level level,
 
 static int
 sshv2_send_command (gftp_request * request, char type, char *command, 
-                    gint32 len)
+                    size_t len)
 {
   char buf[34000];
   gint32 clen;
@@ -846,7 +848,8 @@ sshv2_free_args (char **args)
 static int
 sshv2_connect (gftp_request * request)
 {
-  int version, ret, ssh_use_askpass, sshv2_use_sftp_subsys, fdm;
+  int version, ret, fdm;
+  intptr_t ssh_use_askpass, sshv2_use_sftp_subsys;
   char **args, *tempstr, *p1, p2, *exepath, *ssh2_sftp_path;
   struct servent serv_struct;
   sshv2_params * params;
@@ -1396,8 +1399,8 @@ sshv2_chdir (gftp_request * request, const char *directory)
       sshv2_message_free (&message);
       return (0);
     }
-  else
-    return (sshv2_getcwd (request));
+
+  return (0);
 }
 
 
@@ -1881,7 +1884,7 @@ sshv2_get_file (gftp_request * request, const char *file, int fd,
 {
   sshv2_params * params;
   sshv2_message message;
-  char *tempstr;
+  char *tempstr, *path;
   size_t stlen;
   gint32 num;
   int ret;
@@ -1902,11 +1905,11 @@ sshv2_get_file (gftp_request * request, const char *file, int fd,
     }
   else
     {
-      stlen = strlen (file) + strlen (request->directory) + 1;
+      path = gftp_build_path (request->directory, file, NULL);
+      stlen = strlen (path);
       tempstr = g_malloc (stlen + 16);
-      strcpy (tempstr + 8, request->directory);
-      strcat (tempstr + 8, "/");
-      strcat (tempstr + 8, file);
+      strcpy (tempstr + 8, path);
+      g_free (path);
     }
 
   num = htonl (params->id++);
@@ -1918,7 +1921,8 @@ sshv2_get_file (gftp_request * request, const char *file, int fd,
   num = htonl (SSH_FXF_READ);
   memcpy (tempstr + 8 + stlen, &num, 4);
 
-  memset (tempstr + 12 + stlen, 0, 4);
+  num = 0;
+  memcpy (tempstr + 12 + stlen, &num, 4);
   
   if (sshv2_send_command (request, SSH_FXP_OPEN, tempstr, stlen + 16) < 0)
     {
@@ -1971,7 +1975,7 @@ sshv2_put_file (gftp_request * request, const char *file, int fd,
 {
   sshv2_params * params;
   sshv2_message message;
-  char *tempstr;
+  char *tempstr, *path;
   size_t stlen;
   gint32 num;
   int ret;
@@ -1982,7 +1986,7 @@ sshv2_put_file (gftp_request * request, const char *file, int fd,
   /* fd ignored for this protocol */
 
   params = request->protocol_data;
-  params->offset = 0;
+  params->offset = startsize;
 
   if (*file == '/')
     {
@@ -1992,11 +1996,11 @@ sshv2_put_file (gftp_request * request, const char *file, int fd,
     }
   else
     {
-      stlen = strlen (file) + strlen (request->directory) + 1;
+      path = gftp_build_path (request->directory, file, NULL);
+      stlen = strlen (path);
       tempstr = g_malloc (stlen + 16);
-      strcpy (tempstr + 8, request->directory);
-      strcat (tempstr + 8, "/");
-      strcat (tempstr + 8, file);
+      strcpy (tempstr + 8, path);
+      g_free (path);
     }
 
   num = htonl (params->id++);
@@ -2011,7 +2015,8 @@ sshv2_put_file (gftp_request * request, const char *file, int fd,
     num = htonl (SSH_FXF_WRITE | SSH_FXF_CREAT | SSH_FXF_TRUNC);
   memcpy (tempstr + 8 + stlen, &num, 4);
 
-  memset (tempstr + 12 + stlen, 0, 4);
+  num = 0;
+  memcpy (tempstr + 12 + stlen, &num, 4);
   
   if (sshv2_send_command (request, SSH_FXP_OPEN, tempstr, stlen + 16) < 0)
     {
@@ -2050,7 +2055,7 @@ sshv2_put_file (gftp_request * request, const char *file, int fd,
     }
 
   memset (params->handle, 0, 4);
-  memcpy (params->handle + 4, message.buffer+ 4, message.length - 5);
+  memcpy (params->handle + 4, message.buffer + 4, message.length - 5);
   params->handle_len = message.length - 1;
   sshv2_message_free (&message);
 
@@ -2151,6 +2156,7 @@ sshv2_put_next_file_chunk (gftp_request * request, char *buf, size_t size)
   sshv2_message message;
   char tempstr[32768];
   gint32 num;
+  size_t len;
   int ret;
 
 #ifdef G_HAVE_GINT64
@@ -2167,7 +2173,6 @@ sshv2_put_next_file_chunk (gftp_request * request, char *buf, size_t size)
 
   params = request->protocol_data;
 
-  num = htonl (params->handle_len);
   memcpy (tempstr, params->handle, params->handle_len);
 
   num = htonl (params->id++);
@@ -2186,7 +2191,8 @@ sshv2_put_next_file_chunk (gftp_request * request, char *buf, size_t size)
   memcpy (tempstr + params->handle_len + 8, &num, 4);
   memcpy (tempstr + params->handle_len + 12, buf, size);
   
-  if (sshv2_send_command (request, SSH_FXP_WRITE, tempstr, params->handle_len + size + 12) < 0)
+  len = params->handle_len + size + 12;
+  if (sshv2_send_command (request, SSH_FXP_WRITE, tempstr, len) < 0)
     {
       g_free (tempstr);
       return (GFTP_ERETRYABLE);
@@ -2226,7 +2232,7 @@ sshv2_put_next_file_chunk (gftp_request * request, char *buf, size_t size)
 static int
 sshv2_set_config_options (gftp_request * request)
 {
-  int ssh_need_userpass;
+  intptr_t ssh_need_userpass;
 
   gftp_lookup_request_option (request, "ssh_need_userpass", &ssh_need_userpass);
   request->need_userpass = ssh_need_userpass;
@@ -2261,6 +2267,7 @@ sshv2_init (gftp_request * request)
 
   request->protonum = GFTP_SSHV2_NUM;
   request->init = sshv2_init;
+  request->copy_param_options = NULL;
   request->destroy = sshv2_destroy;
   request->read_function = gftp_fd_read;
   request->write_function = gftp_fd_write;
@@ -2292,7 +2299,6 @@ sshv2_init (gftp_request * request)
   request->need_hostport = 1;
   request->need_userpass = 1;
   request->use_cache = 1;
-  request->use_threads = 1;
   request->always_connected = 0;
   request->protocol_data = g_malloc0 (sizeof (sshv2_params));
   request->server_type = GFTP_DIRTYPE_UNIX;
