@@ -18,11 +18,12 @@
 /*****************************************************************************/
 
 #include "gftp-gtk.h"
-static const char cvsid[] = "$Id: bookmarks.c,v 1.18 2003/08/12 01:05:02 masneyb Exp $";
+static const char cvsid[] = "$Id: bookmarks.c,v 1.27 2004/03/28 23:58:13 masneyb Exp $";
 
-static GtkWidget * bm_hostedit, * bm_portedit, * bm_localdiredit,
-  * bm_remotediredit, * bm_useredit, * bm_passedit, * bm_acctedit, * anon_chk,
-  * bm_pathedit, * bm_protocol, * tree;
+static GtkWidget * bm_dialog = NULL, * bm_hostedit, * bm_portedit,
+                 * bm_localdiredit, * bm_remotediredit, * bm_useredit,
+                 * bm_passedit, * bm_acctedit, * anon_chk, * bm_pathedit,
+                 * bm_protocol, * tree, * edit_bookmarks_dialog;
 static GHashTable * new_bookmarks_htable = NULL;
 static gftp_bookmarks_var * new_bookmarks = NULL;
 static GtkItemFactory * edit_factory;
@@ -31,6 +32,8 @@ static GtkItemFactory * edit_factory;
 void
 run_bookmark (gpointer data)
 {
+  int refresh_local;
+
   if (window1.request->stopable || window2.request->stopable)
     {
       ftp_log (gftp_logging_misc, NULL,
@@ -39,12 +42,15 @@ run_bookmark (gpointer data)
       return;
     }
 
+  if (GFTP_IS_CONNECTED (current_wdata->request))
+    gftpui_disconnect (current_wdata);
+
   if (gftp_parse_bookmark (current_wdata->request, other_wdata->request,
-                           (char *) data) < 0)
+                           (char *) data, &refresh_local) < 0)
     return;
 
-  if (GFTP_IS_CONNECTED (current_wdata->request))
-    disconnect (current_wdata);
+  if (refresh_local)
+    gftpui_refresh (other_wdata);
 
   ftp_connect (current_wdata, current_wdata->request, 1);
 }
@@ -223,6 +229,7 @@ copy_bookmarks (gftp_bookmarks_var * bookmarks)
 
       gftp_copy_local_options (&newentry->local_options_vars,
                                &newentry->local_options_hash,
+                               &newentry->num_local_options_vars,
                                tempentry->local_options_vars,
                                tempentry->num_local_options_vars);
       newentry->num_local_options_vars = tempentry->num_local_options_vars;
@@ -369,7 +376,8 @@ bm_close_dialog (GtkWidget * widget, GtkWidget * dialog)
       new_bookmarks = NULL;
     }
 
-  gtk_widget_destroy (dialog);
+  gtk_widget_destroy (edit_bookmarks_dialog);
+  edit_bookmarks_dialog = NULL;
 }
 
 
@@ -687,10 +695,8 @@ entry_apply_changes (GtkWidget * widget, gftp_bookmarks_var * entry)
       tempchar = *pos;
       *pos = '\0';
     }
-  origpath = newpath = g_strconcat (entry->path, "/",
-                         gtk_entry_get_text (GTK_ENTRY (bm_pathedit)), NULL);
-  remove_double_slashes (newpath);
-  for (; *newpath == '/'; newpath++);
+
+  origpath = newpath = gftp_build_path (entry->path, gtk_entry_get_text (GTK_ENTRY (bm_pathedit)), NULL);
   *pos = tempchar;
 
   str = gtk_entry_get_text (GTK_ENTRY (bm_hostedit));
@@ -752,9 +758,12 @@ entry_apply_changes (GtkWidget * widget, gftp_bookmarks_var * entry)
       while (tempentry != NULL)
 	{
 	  g_hash_table_remove (new_bookmarks_htable, tempentry->path);
-	  tempstr = g_strconcat (newpath, "/", tempentry->path + oldpathlen, 
-                                 NULL);
-	  remove_double_slashes (tempstr);
+
+          if (tempentry->path + oldpathlen == '\0')
+	    tempstr = g_strdup (newpath);
+          else
+	    tempstr = gftp_build_path (newpath, tempentry->path + oldpathlen, NULL);
+
 	  g_free (tempentry->path);
 	  tempentry->path = tempstr;
 	  g_hash_table_insert (new_bookmarks_htable, tempentry->path,
@@ -779,7 +788,18 @@ entry_apply_changes (GtkWidget * widget, gftp_bookmarks_var * entry)
 }
 
 
-#if GTK_MAJOR_VERSION > 1
+#if GTK_MAJOR_VERSION == 1
+
+static void
+entry_close_dialog (void * data)
+{
+  gtk_widget_destroy (bm_dialog);
+  bm_dialog = NULL;
+}
+
+
+#else
+
 static void
 bmedit_action (GtkWidget * widget, gint response, gpointer user_data)
 {
@@ -793,6 +813,7 @@ bmedit_action (GtkWidget * widget, gint response, gpointer user_data)
         /* no break */
       default:
         gtk_widget_destroy (widget);
+        bm_dialog = NULL;
     }
 }   
 #endif
@@ -801,11 +822,17 @@ bmedit_action (GtkWidget * widget, gint response, gpointer user_data)
 static void
 edit_entry (gpointer data)
 {
-  GtkWidget * table, * tempwid, * dialog, * menu, * notebook;
+  GtkWidget * table, * tempwid, * menu, * notebook;
   gftp_bookmarks_var * entry;
   int i, num;
   char *pos;
 
+  if (bm_dialog != NULL)
+    {
+      gtk_widget_grab_focus (bm_dialog);
+      return;
+    }
+
   if (GTK_CLIST (tree)->selection == NULL)
     return;
 
@@ -816,33 +843,34 @@ edit_entry (gpointer data)
     return;
 
 #if GTK_MAJOR_VERSION == 1
-  dialog = gtk_dialog_new ();
-  gtk_window_set_title (GTK_WINDOW (dialog), _("Edit Entry"));
-  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->action_area), 15);
+  bm_dialog = gtk_dialog_new ();
+  gtk_window_set_title (GTK_WINDOW (bm_dialog), _("Edit Entry"));
+  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (bm_dialog)->action_area), 15);
 #else
-  dialog = gtk_dialog_new_with_buttons (_("Edit Entry"), NULL, 0,
-                                        GTK_STOCK_SAVE,
-                                        GTK_RESPONSE_OK,
-                                        GTK_STOCK_CANCEL,
-                                        GTK_RESPONSE_CANCEL,
-                                        GTK_STOCK_APPLY,
-                                        GTK_RESPONSE_APPLY,
-                                        NULL);
+  bm_dialog = gtk_dialog_new_with_buttons (_("Edit Entry"), NULL, 0,
+                                           GTK_STOCK_SAVE,
+                                           GTK_RESPONSE_OK,
+                                           GTK_STOCK_CANCEL,
+                                           GTK_RESPONSE_CANCEL,
+                                           GTK_STOCK_APPLY,
+                                           GTK_RESPONSE_APPLY,
+                                           NULL);
 #endif
-  gtk_window_set_wmclass (GTK_WINDOW (dialog), "Edit Bookmark Entry", "gFTP");
-  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
-  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
-  gtk_widget_realize (dialog);
+  gtk_window_set_wmclass (GTK_WINDOW (bm_dialog), "Edit Bookmark Entry",
+                          "gFTP");
+  gtk_window_set_position (GTK_WINDOW (bm_dialog), GTK_WIN_POS_MOUSE);
+  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (bm_dialog)->vbox), 10);
+  gtk_widget_realize (bm_dialog);
 
   if (gftp_icon != NULL)
     {
-      gdk_window_set_icon (dialog->window, NULL, gftp_icon->pixmap,
+      gdk_window_set_icon (bm_dialog->window, NULL, gftp_icon->pixmap,
                            gftp_icon->bitmap);
-      gdk_window_set_icon_name (dialog->window, gftp_version);
+      gdk_window_set_icon_name (bm_dialog->window, gftp_version);
     }
 
   notebook = gtk_notebook_new ();
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), notebook, TRUE,
+  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (bm_dialog)->vbox), notebook, TRUE,
 		      TRUE, 0);
   gtk_widget_show (notebook);
 
@@ -1012,29 +1040,28 @@ edit_entry (gpointer data)
 
 #if GTK_MAJOR_VERSION == 1
   tempwid = gtk_button_new_with_label (_("OK"));
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid,
+  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (bm_dialog)->action_area), tempwid,
 		      TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
 		      GTK_SIGNAL_FUNC (entry_apply_changes),
 		      (gpointer) entry);
   gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked",
-			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
-			     GTK_OBJECT (dialog));
+			     GTK_SIGNAL_FUNC (entry_close_dialog), NULL);
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_show (tempwid);
 
   tempwid = gtk_button_new_with_label (_("  Cancel  "));
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid,
+  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (bm_dialog)->action_area), tempwid,
 		      TRUE, TRUE, 0);
   gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked",
 			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
-			     GTK_OBJECT (dialog));
+			     GTK_OBJECT (bm_dialog));
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_grab_focus (tempwid);
   gtk_widget_show (tempwid);
 
   tempwid = gtk_button_new_with_label (_("Apply"));
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid,
+  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (bm_dialog)->action_area), tempwid,
 		      TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
 		      GTK_SIGNAL_FUNC (entry_apply_changes),
@@ -1042,13 +1069,13 @@ edit_entry (gpointer data)
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_show (tempwid);
 #else
-  g_signal_connect (GTK_OBJECT (dialog), "response",
+  g_signal_connect (GTK_OBJECT (bm_dialog), "response",
                     G_CALLBACK (bmedit_action), (gpointer) entry);
 #endif
 
   gftp_gtk_setup_bookmark_options (notebook, entry);
 
-  gtk_widget_show (dialog);
+  gtk_widget_show (bm_dialog);
 }
 
 
@@ -1072,6 +1099,14 @@ bm_enter (GtkWidget * widget, GdkEventKey * event, gpointer data)
 }
 
 
+static inline gboolean 
+move_possible (GtkCTree *ctree, GtkCTreeNode *child, GtkCTreeNode *parent,
+               GtkCTreeNode *sibling)
+{
+  return (parent != NULL && sibling != NULL);
+}
+
+
 static void
 after_move (GtkCTree * ctree, GtkCTreeNode * child, GtkCTreeNode * parent,
 	    GtkCTreeNode * sibling, gpointer data)
@@ -1172,8 +1207,8 @@ bm_dblclick (GtkWidget * widget, GdkEventButton * event, gpointer data)
 void
 edit_bookmarks (gpointer data)
 {
-  GtkWidget * dialog, * scroll;
   GtkItemFactory * ifactory;
+  GtkWidget * scroll;
   GtkItemFactoryEntry menu_items[] = {
     {N_("/_File"), NULL, 0, 0, "<Branch>"},
     {N_("/File/tearoff"), NULL, 0, 0, "<Tearoff>"},
@@ -1188,37 +1223,46 @@ edit_bookmarks (gpointer data)
   GtkWidget * tempwid;
 #endif
 
+  if (edit_bookmarks_dialog != NULL)
+    {
+      gtk_widget_grab_focus (edit_bookmarks_dialog);
+      return;
+    }
+
   new_bookmarks = copy_bookmarks (gftp_bookmarks);
   new_bookmarks_htable = build_bookmarks_hash_table (new_bookmarks);
 
 #if GTK_MAJOR_VERSION == 1
-  dialog = gtk_dialog_new ();
-  gtk_window_set_title (GTK_WINDOW (dialog), _("Edit Bookmarks"));
-  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->action_area), 15);
+  edit_bookmarks_dialog = gtk_dialog_new ();
+  gtk_window_set_title (GTK_WINDOW (edit_bookmarks_dialog),
+                        _("Edit Bookmarks"));
+  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (edit_bookmarks_dialog)->action_area), 15);
 #else
-  dialog = gtk_dialog_new_with_buttons (_("Edit Bookmarks"), NULL, 0,
-                                        GTK_STOCK_SAVE,
-                                        GTK_RESPONSE_OK,
-                                        GTK_STOCK_CANCEL,
-                                        GTK_RESPONSE_CANCEL,
-                                        NULL);
+  edit_bookmarks_dialog = gtk_dialog_new_with_buttons (_("Edit Bookmarks"),
+                                                       NULL, 0, GTK_STOCK_SAVE,
+                                                       GTK_RESPONSE_OK,
+                                                       GTK_STOCK_CANCEL,
+                                                       GTK_RESPONSE_CANCEL,
+                                                       NULL);
 #endif
-  gtk_window_set_wmclass (GTK_WINDOW(dialog), "Edit Bookmarks", "gFTP");
-  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
-  gtk_widget_realize (dialog);
+  gtk_window_set_wmclass (GTK_WINDOW(edit_bookmarks_dialog), "Edit Bookmarks",
+                          "gFTP");
+  gtk_window_set_position (GTK_WINDOW (edit_bookmarks_dialog),
+                           GTK_WIN_POS_MOUSE);
+  gtk_widget_realize (edit_bookmarks_dialog);
 
   if (gftp_icon != NULL)
     {
-      gdk_window_set_icon (dialog->window, NULL, gftp_icon->pixmap,
-                           gftp_icon->bitmap);
-      gdk_window_set_icon_name (dialog->window, gftp_version);
+      gdk_window_set_icon (edit_bookmarks_dialog->window, NULL,
+                           gftp_icon->pixmap, gftp_icon->bitmap);
+      gdk_window_set_icon_name (edit_bookmarks_dialog->window, gftp_version);
     }
 
   ifactory = item_factory_new (GTK_TYPE_MENU_BAR, "<bookmarks>", NULL, NULL);
   create_item_factory (ifactory, 7, menu_items, NULL);
-  create_item_factory (ifactory, 1, menu_items + 7, dialog);
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), ifactory->widget,
-		      FALSE, FALSE, 0);
+  create_item_factory (ifactory, 1, menu_items + 7, edit_bookmarks_dialog);
+  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (edit_bookmarks_dialog)->vbox),
+                      ifactory->widget, FALSE, FALSE, 0);
   gtk_widget_show (ifactory->widget);
 
   scroll = gtk_scrolled_window_new (NULL, NULL);
@@ -1226,14 +1270,14 @@ edit_bookmarks (gpointer data)
 				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_widget_set_size_request (scroll, 150, 200);
 
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), scroll, TRUE, TRUE,
-		      0);
+  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (edit_bookmarks_dialog)->vbox),
+                      scroll, TRUE, TRUE, 0);
   gtk_container_border_width (GTK_CONTAINER (scroll), 3);
   gtk_widget_show (scroll);
 
   edit_factory = item_factory_new (GTK_TYPE_MENU, "<edit_bookmark>", NULL, "/File");
 
-  create_item_factory (edit_factory, 6, menu_items + 2, dialog);
+  create_item_factory (edit_factory, 6, menu_items + 2, edit_bookmarks_dialog);
 
   tree = gtk_ctree_new (1, 0);
   gtk_clist_set_selection_mode (GTK_CLIST (tree), GTK_SELECTION_BROWSE);
@@ -1245,33 +1289,36 @@ edit_bookmarks (gpointer data)
 			    GTK_SIGNAL_FUNC (after_move), NULL);
   gtk_signal_connect_after (GTK_OBJECT (tree), "button_press_event",
 			    GTK_SIGNAL_FUNC (bm_dblclick), (gpointer) tree);
+  gtk_ctree_set_drag_compare_func (GTK_CTREE(tree), &move_possible);
   gtk_widget_show (tree);
 
 #if GTK_MAJOR_VERSION == 1
   tempwid = gtk_button_new_with_label (_("OK"));
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid,
-		      TRUE, TRUE, 0);
+  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (edit_bookmarks_dialog)->action_area),
+                      tempwid, TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
 		      GTK_SIGNAL_FUNC (bm_apply_changes), NULL);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
-		      GTK_SIGNAL_FUNC (bm_close_dialog), (gpointer) dialog);
+		      GTK_SIGNAL_FUNC (bm_close_dialog),
+                      (gpointer) edit_bookmarks_dialog);
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_show (tempwid);
 
   tempwid = gtk_button_new_with_label (_("  Cancel  "));
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid,
-		      TRUE, TRUE, 0);
+  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (edit_bookmarks_dialog)->action_area),
+                      tempwid, TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
-		      GTK_SIGNAL_FUNC (bm_close_dialog), (gpointer) dialog);
+		      GTK_SIGNAL_FUNC (bm_close_dialog),
+                      (gpointer) edit_bookmarks_dialog);
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_grab_focus (tempwid);
   gtk_widget_show (tempwid);
 #else
-  g_signal_connect (GTK_OBJECT (dialog), "response",
+  g_signal_connect (GTK_OBJECT (edit_bookmarks_dialog), "response",
                     G_CALLBACK (editbm_action), NULL);
 #endif
 
-  gtk_widget_show (dialog);
+  gtk_widget_show (edit_bookmarks_dialog);
 
   build_bookmarks_tree ();
 }
