@@ -18,14 +18,15 @@
 /*****************************************************************************/
 
 #include "gftp-gtk.h"
-static const char cvsid[] = "$Id: dnd.c,v 1.16 2003/08/12 01:05:02 masneyb Exp $";
+static const char cvsid[] = "$Id: dnd.c,v 1.22 2004/02/01 22:24:47 masneyb Exp $";
 
 
 static int
-dnd_remote_file (char *url, gftp_window_data * wdata)
+dnd_remote_file (gftp_window_data * wdata, GList ** trans_list, char *url)
 {
   gftp_window_data * other_wdata, * fromwdata;
   gftp_request * current_ftpdata;
+  gftp_transfer * tdata;
   gftp_file * newfle;
   GList * templist;
   char *pos;
@@ -61,11 +62,12 @@ dnd_remote_file (char *url, gftp_window_data * wdata)
   if (compare_request (current_ftpdata, wdata->request, 1))
     {
       gftp_request_destroy (current_ftpdata, 1);
+      free_fdata (newfle);
       return (0);
     }
 
   if (other_wdata != NULL && 
-      compare_request (current_ftpdata, other_wdata->request, 1))
+      compare_request (current_ftpdata, other_wdata->request, 0))
     {
       if (other_wdata->request->password != NULL)
         gftp_set_password (current_ftpdata, other_wdata->request->password);
@@ -80,14 +82,30 @@ dnd_remote_file (char *url, gftp_window_data * wdata)
   *(pos - 1) = '\0';
   
   newfle->destfile = gftp_build_path (wdata->request->directory, pos, NULL);
-  templist = g_malloc0 (sizeof (*templist));
-  templist->data = newfle;
-  templist->next = NULL;
-  add_file_transfer (current_ftpdata, wdata->request, fromwdata,
-                     wdata, templist, 1);
-  gftp_request_destroy (current_ftpdata, 1);
 
-  /* FIXME  resume files and download entire directories */
+  for (templist = *trans_list; templist != NULL; templist = templist->next)
+    {
+      tdata = templist->data;
+      if (compare_request (current_ftpdata, tdata->fromreq, 0))
+        break;
+    }
+
+  if (templist == NULL)
+    {
+      tdata = gftp_tdata_new ();
+
+      tdata->fromreq = current_ftpdata;
+      tdata->fromwdata = fromwdata;
+
+      tdata->toreq = wdata->request;
+      tdata->towdata = wdata;
+
+      *trans_list = g_list_append (*trans_list, tdata);
+    }
+  else
+    gftp_request_destroy (current_ftpdata, 1);
+
+  tdata->files = g_list_append (tdata->files, newfle);
 
   return (1);
 }
@@ -100,14 +118,12 @@ openurl_get_drag_data (GtkWidget * widget, GdkDragContext * context, gint x,
 {
   if ((selection_data->length >= 0) && (selection_data->format == 8)) 
     {
+      if (GFTP_IS_CONNECTED (current_wdata->request))
+        gftpui_disconnect (current_wdata);
+
       if (gftp_parse_url (current_wdata->request, 
                           (char *) selection_data->data) == 0)
-        {
-          if (GFTP_IS_CONNECTED (current_wdata->request))
-            disconnect (current_wdata);
-
-          ftp_connect (current_wdata, current_wdata->request, 1);
-        }
+        ftp_connect (current_wdata, current_wdata->request, 1);
     }
 }
 
@@ -118,7 +134,7 @@ listbox_drag (GtkWidget * widget, GdkDragContext * context,
 	      gpointer data)
 {
   GList * templist, * filelist;
-  char *tempstr, *str, *pos;
+  char *tempstr, *str, *df;
   gftp_window_data * wdata;
   size_t totlen, oldlen;
   gftp_file * tempfle;
@@ -142,40 +158,32 @@ listbox_drag (GtkWidget * widget, GdkDragContext * context,
         continue;
 
       oldlen = totlen;
+      df = gftp_build_path (wdata->request->directory, tempfle->file, NULL);
+
       if (wdata->request->hostname == NULL || 
           wdata->request->protonum == GFTP_LOCAL_NUM)
         {
-          tempstr = g_strdup_printf ("%s://%s/%s ", 
-                                 wdata->request->url_prefix,
-                                 wdata->request->directory, 
-                                 tempfle->file);
+          tempstr = g_strdup_printf ("%s://%s ", 
+                                 wdata->request->url_prefix, df);
         }
-      else if (wdata->request->username == NULL 
-               || *wdata->request->username == '\0')
+      else if (wdata->request->username == NULL || 
+               *wdata->request->username == '\0')
         {
-          tempstr = g_strdup_printf ("%s://%s:%d%s/%s ", 
+          tempstr = g_strdup_printf ("%s://%s:%d%s ", 
                                  wdata->request->url_prefix,
                                  wdata->request->hostname,
-                                 wdata->request->port,
-                                 wdata->request->directory, 
-                                 tempfle->file);
+                                 wdata->request->port, df);
         }
       else
         {
-          tempstr = g_strdup_printf ("%s://%s@%s:%d%s/%s ", 
+          tempstr = g_strdup_printf ("%s://%s@%s:%d%s ", 
                                  wdata->request->url_prefix,
                                  wdata->request->username, 
                                  wdata->request->hostname,
-                                 wdata->request->port,
-                                 wdata->request->directory, 
-                                 tempfle->file);
+                                 wdata->request->port, df);
         }
 
-      if ((pos = strchr (tempstr, ':')) != NULL)
-        pos += 3;
-      else
-        pos = tempstr;
-      remove_double_slashes (pos);
+      g_free (df);
 
       /* Note, I am allocating memory for this byte above. Note the extra space
          at the end of the g_strdup_printf() format argument */
@@ -214,14 +222,18 @@ listbox_get_drag_data (GtkWidget * widget, GdkDragContext * context, gint x,
 		       gint y, GtkSelectionData * selection_data, guint info,
 		       guint32 clk_time, gpointer data)
 {
-  char *newpos, *oldpos, tempchar;
+  char *newpos, *oldpos, *tempstr;
+  GList * trans_list, * templist;
   gftp_window_data * wdata;
+  gftp_transfer * tdata;
   int finish_drag;
+  size_t len;
 
   wdata = data;   
   if (!check_status (_("Drag-N-Drop"), wdata, 1, 0, 0, 1)) 
     return;
 
+  trans_list = NULL;
   finish_drag = 0;
   if ((selection_data->length >= 0) && (selection_data->format == 8)) 
     {
@@ -229,17 +241,54 @@ listbox_get_drag_data (GtkWidget * widget, GdkDragContext * context, gint x,
       while ((newpos = strchr (oldpos, '\n')) || 
              (newpos = strchr (oldpos, '\0'))) 
         {
-          tempchar = *newpos;
-          *newpos = '\0';
-          ftp_log (gftp_logging_misc, NULL, _("Received URL %s\n"), oldpos);
-          if (dnd_remote_file (oldpos, wdata))
+          len = newpos - oldpos;
+          if (oldpos[len - 1] == '\r')
+            len--;
+
+          if (len == 0)
+            break;
+
+          tempstr = g_malloc (len + 1);
+          memcpy (tempstr, oldpos, len);
+          tempstr[len] = '\0';
+
+          ftp_log (gftp_logging_misc, NULL, _("Received URL %s\n"), tempstr);
+
+          if (dnd_remote_file (wdata, &trans_list, tempstr))
             finish_drag = 1;
-         
+
+          g_free (tempstr);
+
           if (*newpos == '\0') 
             break;
+
           oldpos = newpos + 1;
         }
     }
 
   gtk_drag_finish (context, finish_drag, FALSE, clk_time);
+
+  for (templist = trans_list; templist != NULL; templist = templist->next)
+    {
+      tdata = templist->data;
+
+      if (tdata->files == NULL || gftp_get_all_subdirs (tdata, NULL) != 0)
+        {
+          g_free (tdata);
+          continue;
+        }
+
+      if (tdata->files == NULL)
+        {
+          g_free (tdata);
+          continue;
+        }
+
+      gftpui_common_add_file_transfer (tdata->fromreq, tdata->toreq,
+                                       tdata->fromwdata, tdata->towdata,
+                                       tdata->files);
+      g_free (tdata);
+    }
+
+  g_list_free (trans_list);
 }
