@@ -18,15 +18,15 @@
 /*****************************************************************************/
 
 #include "gftp-gtk.h"
-static const char cvsid[] = "$Id: gftp-gtk.c,v 1.31 2003/08/12 01:05:02 masneyb Exp $";
+static const char cvsid[] = "$Id: gftp-gtk.c,v 1.45 2004/03/21 23:56:18 masneyb Exp $";
 
 static GtkItemFactory *log_factory, *dl_factory;
 static GtkWidget * local_frame, * remote_frame, * log_table, * transfer_scroll,
-                 * openurl_btn;
+                 * openurl_btn, * gftpui_command_toolbar;
 
 gftp_window_data window1, window2, *other_wdata, *current_wdata;
 GtkWidget * stop_btn, * hostedit, * useredit, * passedit, * portedit, * logwdw,
-          * dlwdw, * protocol_menu, * optionmenu;
+          * dlwdw, * protocol_menu, * optionmenu, * gftpui_command_widget;
 GtkAdjustment * logwdw_vadj;
 #if GTK_MAJOR_VERSION > 1
 GtkTextMark * logwdw_textmark;
@@ -35,11 +35,8 @@ int local_start, remote_start, trans_start, log_start, tools_start;
 GHashTable * graphic_hash_table = NULL;
 GtkItemFactoryEntry * menus = NULL;
 GtkItemFactory * factory = NULL;
-pthread_mutex_t transfer_mutex = PTHREAD_MUTEX_INITIALIZER;
 pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER;
 gftp_graphic * gftp_icon;
-sigjmp_buf jmp_environment;
-volatile int use_jmp_environment = 0;
 pthread_t main_thread_id;
 GList * viewedit_processes = NULL;
 volatile sig_atomic_t viewedit_process_done;
@@ -127,13 +124,6 @@ destroy (GtkWidget * widget, gpointer data)
 }
 
 
-static RETSIGTYPE
-sig_child (int signo)
-{
-  viewedit_process_done = 1;
-}
-
-
 static void
 menu_exit (GtkWidget * widget, gpointer data)
 {
@@ -143,16 +133,90 @@ menu_exit (GtkWidget * widget, gpointer data)
 
 
 static void
-chfunc (gpointer data)
+change_setting (gftp_window_data * wdata, int menuitem, GtkWidget * checkmenu)
 {
-  chdir_dialog (data);
+  switch (menuitem)
+    {
+    case 1:
+      gftp_set_global_option ("ascii_transfers", GINT_TO_POINTER(1));
+      break;
+    case 2:
+      gftp_set_global_option ("ascii_transfers", GINT_TO_POINTER(0));
+      break;
+    case 3:
+      current_wdata = &window1;
+      other_wdata = &window2;
+      if (wdata->request)
+        update_window_info ();
+      break;
+    case 4:
+      current_wdata = &window2;
+      other_wdata = &window1;
+      if (wdata->request)
+        update_window_info ();
+      break;
+    }
+}
+
+
+static void
+_gftpui_gtk_do_openurl (gftp_window_data * wdata, gftp_dialog_data * ddata)
+{
+  const char *tempstr;
+  char *buf;
+
+  tempstr = gtk_entry_get_text (GTK_ENTRY (ddata->edit));
+  if (tempstr != NULL && *tempstr != '\0')
+    {
+      buf = g_strdup (tempstr);
+      destroy_dialog (ddata);
+      gftpui_common_cmd_open (wdata, wdata->request, NULL, NULL, buf);
+      g_free (buf);
+    }
+}
+
+
+static void
+openurl_dialog (gpointer data)
+{
+  gftp_window_data * wdata;
+
+  wdata = data;
+  MakeEditDialog (_("Connect via URL"), _("Enter a URL to connect to"),
+                  NULL, 1, NULL, gftp_dialog_button_connect,
+                  _gftpui_gtk_do_openurl, wdata,
+                  NULL, NULL);
+}
+
+
+static void
+tb_openurl_dialog (gpointer data)
+{
+  const char *edttxt;
+
+  if (current_wdata->request->stopable)
+    {
+      ftp_log (gftp_logging_misc, NULL,
+               _("%s: Please hit the stop button first to do anything else\n"),
+               _("OpenURL"));
+      return;
+    }
+
+  edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (hostedit)->entry));
+
+  if (GFTP_IS_CONNECTED (current_wdata->request))
+    gftpui_disconnect (current_wdata);
+  else if (edttxt != NULL && *edttxt != '\0')
+    toolbar_hostedit (NULL, NULL);
+  else
+    openurl_dialog (current_wdata);
 }
 
 
 static GtkWidget *
 CreateMenus (GtkWidget * parent)
 {
-  int local_len, remote_len, len, i, trans_len, log_len, tools_len,
+  intptr_t local_len, remote_len, len, i, trans_len, log_len, tools_len,
       ascii_transfers;
   GtkAccelGroup *accel_group;
   GtkWidget * tempwid;
@@ -172,7 +236,7 @@ CreateMenus (GtkWidget * parent)
     {N_("/_Local"), NULL, 0, 0, MN_("<Branch>")},
     {N_("/Local/tearoff"), NULL, 0, 0, MN_("<Tearoff>")},
     {N_("/Local/Open _URL..."), NULL, openurl_dialog, 0, MS_(GTK_STOCK_OPEN)},
-    {N_("/Local/Disconnect"), NULL, disconnect, 0, MS_(GTK_STOCK_CLOSE)},
+    {N_("/Local/Disconnect"), NULL, gftpui_disconnect, 0, MS_(GTK_STOCK_CLOSE)},
     {N_("/Local/sep"), NULL, 0, 0, MN_("<Separator>")},
     {N_("/Local/Change Filespec..."), NULL, change_filespec, 0, MN_(NULL)},
     {N_("/Local/Show selected"), NULL, show_selected, 0, MN_(NULL)},
@@ -181,20 +245,20 @@ CreateMenus (GtkWidget * parent)
     {N_("/Local/Deselect All"), NULL, deselectall, 0, MN_(NULL)},
     {N_("/Local/sep"), NULL, 0, 0, MN_("<Separator>")},
     {N_("/Local/Save Directory Listing..."), NULL, save_directory_listing, 0, MN_(NULL)},
-    {N_("/Local/Send SITE Command..."), NULL, site_dialog, 0, MN_(NULL)},
-    {N_("/Local/Change Directory"), NULL, chfunc, 0, MN_(NULL)},
+    {N_("/Local/Send SITE Command..."), NULL, gftpui_site_dialog, 0, MN_(NULL)},
+    {N_("/Local/Change Directory"), NULL, gftpui_chdir_dialog, 0, MN_(NULL)},
     {N_("/Local/Chmod..."), NULL, chmod_dialog, 0, MN_(NULL)},
-    {N_("/Local/Make Directory..."), NULL, mkdir_dialog, 0, MN_(NULL)},
-    {N_("/Local/Rename..."), NULL, rename_dialog, 0, MN_(NULL)},
+    {N_("/Local/Make Directory..."), NULL, gftpui_mkdir_dialog, 0, MN_(NULL)},
+    {N_("/Local/Rename..."), NULL, gftpui_rename_dialog, 0, MN_(NULL)},
     {N_("/Local/Delete..."), NULL, delete_dialog, 0, MN_(NULL)},
     {N_("/Local/Edit..."), NULL, edit_dialog, 0, MN_(NULL)},
     {N_("/Local/View..."), NULL, view_dialog, 0, MN_(NULL)},
-    {N_("/Local/Refresh"), NULL, refresh, 0, MS_(GTK_STOCK_REFRESH)},
+    {N_("/Local/Refresh"), NULL, gftpui_refresh, 0, MS_(GTK_STOCK_REFRESH)},
     {N_("/_Remote"), NULL, 0, 0, MN_("<Branch>")},
     {N_("/Remote/tearoff"), NULL, 0, 0, MN_("<Tearoff>")},
     {N_("/Remote/Open _URL..."), "<control>U", openurl_dialog, 0,
 	MS_(GTK_STOCK_OPEN)},
-    {N_("/Remote/Disconnect"), "<control>D", disconnect, 0,
+    {N_("/Remote/Disconnect"), "<control>D", gftpui_disconnect, 0,
 	MS_(GTK_STOCK_CLOSE)},
     {N_("/Remote/sep"), NULL, 0, 0, MN_("<Separator>")},
     {N_("/Remote/Change Filespec..."), NULL, change_filespec, 0, MN_(NULL)},
@@ -204,15 +268,15 @@ CreateMenus (GtkWidget * parent)
     {N_("/Remote/Deselect All"), NULL, deselectall, 0, MN_(NULL)},
     {N_("/Remote/sep"), NULL, 0, 0, MN_("<Separator>")},
     {N_("/Remote/Save Directory Listing..."), NULL, save_directory_listing, 0, MN_(NULL)},
-    {N_("/Remote/Send SITE Command..."), NULL, site_dialog, 0, MN_(NULL)},
-    {N_("/Remote/Change Directory"), NULL, chfunc, 0, MN_(NULL)},
+    {N_("/Remote/Send SITE Command..."), NULL, gftpui_site_dialog, 0, MN_(NULL)},
+    {N_("/Remote/Change Directory"), NULL, gftpui_chdir_dialog, 0, MN_(NULL)},
     {N_("/Remote/Chmod..."), NULL, chmod_dialog, 0, MN_(NULL)},
-    {N_("/Remote/Make Directory..."), NULL, mkdir_dialog, 0, MN_(NULL)},
-    {N_("/Remote/Rename..."), NULL, rename_dialog, 0, MN_(NULL)},
+    {N_("/Remote/Make Directory..."), NULL, gftpui_mkdir_dialog, 0, MN_(NULL)},
+    {N_("/Remote/Rename..."), NULL, gftpui_rename_dialog, 0, MN_(NULL)},
     {N_("/Remote/Delete..."), NULL, delete_dialog, 0, MN_(NULL)},
     {N_("/Remote/Edit..."), NULL, edit_dialog, 0, MN_(NULL)},
     {N_("/Remote/View..."), NULL, view_dialog, 0, MN_(NULL)},
-    {N_("/Remote/Refresh"), NULL, refresh, 0, MS_(GTK_STOCK_REFRESH)},
+    {N_("/Remote/Refresh"), NULL, gftpui_refresh, 0, MS_(GTK_STOCK_REFRESH)},
     {N_("/_Bookmarks"), NULL, 0, 0, MN_("<Branch>")},
     {N_("/Bookmarks/tearoff"), NULL, 0, 0, MN_("<Tearoff>")},
     {N_("/Bookmarks/Add bookmark"), "<control>A", add_bookmark, 0,
@@ -244,7 +308,7 @@ CreateMenus (GtkWidget * parent)
     {N_("/Tools/tearoff"), NULL, 0, 0, MN_("<Tearoff>")},
     {N_("/Tools/Compare Windows"), NULL, compare_windows, 0, MN_(NULL)},
     {N_("/Tools/Clear Cache"), NULL, clear_cache, 0, MS_(GTK_STOCK_CLEAR)},
-    {N_("/_Help"), NULL, 0, 0, MN_("<LastBranch>")},
+    {N_("/_Help"), NULL, 0, 0, MN_("<Branch>")},
     {N_("/Help/tearoff"), NULL, 0, 0, MN_("<Tearoff>")},
     {N_("/Help/About..."), NULL, about_dialog, 0, MS_(GTK_STOCK_HELP)}
   };
@@ -334,7 +398,7 @@ CreateMenus (GtkWidget * parent)
 
 
 static GtkWidget *
-CreateToolbar (GtkWidget * parent)
+CreateConnectToolbar (GtkWidget * parent)
 {
   const GtkTargetEntry possible_types[] = {
     {"STRING", 0, 0},
@@ -470,10 +534,35 @@ CreateToolbar (GtkWidget * parent)
   gtk_container_border_width (GTK_CONTAINER (stop_btn), 1);
   gtk_box_pack_start (GTK_BOX (box), stop_btn, FALSE, FALSE, 0);
 
+  gtk_widget_grab_focus (GTK_COMBO (hostedit)->entry);
+
   return (toolbar);
 }
 
 
+static GtkWidget *
+CreateCommandToolbar (void)
+{
+  GtkWidget * box, * tempwid;
+
+  gftpui_command_toolbar = gtk_handle_box_new ();
+
+  box = gtk_hbox_new (FALSE, 4);
+  gtk_container_add (GTK_CONTAINER (gftpui_command_toolbar), box);
+  gtk_container_border_width (GTK_CONTAINER (box), 5);
+
+  tempwid = gtk_label_new (_("Command: "));
+  gtk_box_pack_start (GTK_BOX (box), tempwid, FALSE, FALSE, 0);
+
+  gftpui_command_widget = gtk_entry_new ();
+  gtk_box_pack_start (GTK_BOX (box), gftpui_command_widget, TRUE, TRUE, 0);
+  gtk_signal_connect (GTK_OBJECT (gftpui_command_widget), "activate",
+		      GTK_SIGNAL_FUNC (gftpui_run_command), NULL);
+
+  return (gftpui_command_toolbar);
+}
+
+
 static void
 setup_column (GtkWidget * listbox, int column, int width)
 {
@@ -489,9 +578,11 @@ setup_column (GtkWidget * listbox, int column, int width)
 static void
 list_doaction (gftp_window_data * wdata)
 {
-  int num, dir, success, list_dblclk_action;
+  intptr_t list_dblclk_action;
   GList *templist, *filelist;
+  int num, dir, success;
   gftp_file *tempfle;
+  char *directory;
 
   gftp_lookup_request_option (wdata->request, "list_dblclk_action", 
                               &list_dblclk_action);
@@ -503,10 +594,18 @@ list_doaction (gftp_window_data * wdata)
   tempfle = (gftp_file *) filelist->data;
 
   dir = tempfle->isdir;
-  success = 0;
+
+  if (check_reconnect (wdata) < 0) 
+    return;
 
   if (tempfle->islink || tempfle->isdir)
-    success = chdir_dialog (wdata);
+    {
+      directory = gftp_build_path (wdata->request->directory, tempfle->file, NULL);
+      success = gftpui_run_chdir (wdata, directory);
+      g_free (directory);
+    }
+  else
+    success = 0;
 
   if (!dir && !success)
     {
@@ -591,7 +690,7 @@ CreateFTPWindow (gftp_window_data * wdata)
     {"application/x-rootwin-drop", 0, 1}
   };
   GtkWidget *box, *scroll_list, *parent;
-  int listbox_file_height, colwidth;
+  intptr_t listbox_file_height, colwidth;
   char *titles[7], tempstr[50];
 
   titles[0] = "";
@@ -718,7 +817,7 @@ CreateFTPWindows (GtkWidget * ui)
             *button;
   gftp_config_list_vars * tmplistvar;
   char *dltitles[2];
-  int tmplookup;
+  intptr_t tmplookup;
 #if GTK_MAJOR_VERSION > 1
   GtkTextBuffer * textbuf;
   GtkTextIter iter;
@@ -742,7 +841,10 @@ CreateFTPWindows (GtkWidget * ui)
   tempwid = CreateMenus (ui);
   gtk_box_pack_start (GTK_BOX (mainvbox), tempwid, FALSE, FALSE, 0);
 
-  tempwid = CreateToolbar (ui);
+  tempwid = CreateConnectToolbar (ui);
+  gtk_box_pack_start (GTK_BOX (mainvbox), tempwid, FALSE, FALSE, 0);
+
+  tempwid = CreateCommandToolbar ();
   gtk_box_pack_start (GTK_BOX (mainvbox), tempwid, FALSE, FALSE, 0);
 
   winpane = gtk_hpaned_new ();
@@ -880,6 +982,7 @@ CreateFTPWindows (GtkWidget * ui)
   gtk_box_pack_start (GTK_BOX (mainvbox), logpane, TRUE, TRUE, 0);
 
   gtk_widget_show_all (mainvbox);
+  gftpui_show_or_hide_command ();
   return (mainvbox);
 }
 
@@ -915,7 +1018,7 @@ toolbar_hostedit (GtkWidget * widget, gpointer data)
     }
 
   if (GFTP_IS_CONNECTED (current_wdata->request))
-    disconnect (current_wdata);
+    gftpui_disconnect (current_wdata);
 
   tempwid = gtk_menu_get_active (GTK_MENU (protocol_menu));
   num = GPOINTER_TO_INT (gtk_object_get_user_data (GTK_OBJECT (tempwid)));
@@ -923,7 +1026,15 @@ toolbar_hostedit (GtkWidget * widget, gpointer data)
   if (init (current_wdata->request) < 0)
     return;
  
-  gftp_set_hostname (current_wdata->request, gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (hostedit)->entry)));
+  txt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (hostedit)->entry));
+  if (strchr (txt, '/') != NULL) 
+    {
+      /* The user entered a URL in the host box... */
+      gftpui_common_cmd_open (current_wdata, current_wdata->request, NULL, NULL, txt);
+      return;
+    }
+
+  gftp_set_hostname (current_wdata->request, txt);
   if (current_wdata->request->hostname == NULL)
     return;
   alltrim (current_wdata->request->hostname);
@@ -981,7 +1092,8 @@ void
 sortrows (GtkCList * clist, gint column, gpointer data)
 {
   char sortcol_name[25], sortasds_name[25];
-  int swap_col, sortcol, sortasds;
+  int swap_col;
+  intptr_t sortcol, sortasds;
   gftp_window_data * wdata;
   GtkWidget * sort_wid;
   GList * templist;
@@ -1079,6 +1191,7 @@ gftp_gtk_config_file_read_color (char *str, gftp_config_vars * cv, int line)
   color->red = strtol (red, NULL, 16);
   color->green = strtol (green, NULL, 16);
   color->blue = strtol (blue, NULL, 16);
+
   g_free (red);
   g_free (green);
   g_free (blue);
@@ -1091,12 +1204,13 @@ gftp_gtk_config_file_read_color (char *str, gftp_config_vars * cv, int line)
 
 
 static int
-gftp_gtk_config_file_write_color (gftp_config_vars * cv, FILE * fd, int to_config_file)
+gftp_gtk_config_file_write_color (gftp_config_vars * cv, char *buf,
+                                  size_t buflen, int to_config_file)
 {
   GdkColor * color;
 
   color = cv->value;
-  fprintf (fd, "%x:%x:%x", color->red, color->green, color->blue);
+  g_snprintf (buf, buflen, "%x:%x:%x", color->red, color->green, color->blue);
   return (0);
 }
 
@@ -1104,39 +1218,23 @@ gftp_gtk_config_file_write_color (gftp_config_vars * cv, FILE * fd, int to_confi
 int
 main (int argc, char **argv)
 {
-  char *startup_directory;
   GtkWidget *window, *ui;
 
-#ifdef HAVE_GETTEXT
-  setlocale (LC_ALL, "");
-  bindtextdomain ("gftp", LOCALE_DIR);
-#if GTK_MAJOR_VERSION > 1
-  bind_textdomain_codeset ("gftp", "UTF-8");
-#endif
-  textdomain ("gftp");
-#endif
-
-  g_thread_init (NULL);
-  main_thread_id = pthread_self ();
-  gtk_set_locale ();
-  gtk_init (&argc, &argv);
-
-  signal (SIGCHLD, sig_child);
-  signal (SIGPIPE, SIG_IGN);
-  signal (SIGALRM, signal_handler);
-  signal (SIGINT, signal_handler);
-
-  graphic_hash_table = g_hash_table_new (string_hash_function, string_hash_compare);
- 
   /* We override the read color functions because we are using a GdkColor 
      structures to store the color. If I put this in lib/config_file.c, then 
      the core library would be dependant on Gtk+ being present */
   gftp_option_types[gftp_option_type_color].read_function = gftp_gtk_config_file_read_color;
   gftp_option_types[gftp_option_type_color].write_function = gftp_gtk_config_file_write_color;
 
-  gftp_read_config_file (SHARE_DIR);
-  if (gftp_parse_command_line (&argc, &argv) != 0)
-    exit (0);
+  gftpui_common_init (&argc, &argv, ftp_log);
+
+  g_thread_init (NULL);
+  main_thread_id = pthread_self ();
+  gtk_set_locale ();
+  gtk_init (&argc, &argv);
+
+  graphic_hash_table = g_hash_table_new (string_hash_function,
+                                         string_hash_compare);
 
   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   gtk_signal_connect (GTK_OBJECT (window), "delete_event",
@@ -1163,22 +1261,12 @@ main (int argc, char **argv)
   gtk_container_add (GTK_CONTAINER (window), ui);
   gtk_widget_show (window);
 
-  ftp_log (gftp_logging_misc, NULL,
-	   "%s, Copyright (C) 1998-2003 Brian Masney <", gftp_version);
-  ftp_log (gftp_logging_recv, NULL, "masneyb@gftp.org");
-  ftp_log (gftp_logging_misc, NULL,
-	   _(">. If you have any questions, comments, or suggestions about this program, please feel free to email them to me. You can always find out the latest news about gFTP from my website at http://www.gftp.org/\n"));
-  ftp_log (gftp_logging_misc, NULL,
-	   _("gFTP comes with ABSOLUTELY NO WARRANTY; for details, see the COPYING file. This is free software, and you are welcome to redistribute it under certain conditions; for details, see the COPYING file\n"));
+  gftpui_common_about (ftp_log, NULL);
 
   gtk_timeout_add (1000, update_downloads, NULL);
   if (gftp_protocols[GFTP_LOCAL_NUM].init (window1.request) == 0)
     {
-      gftp_lookup_request_option (window1.request, "startup_directory", 
-                                  &startup_directory);
-      if (*startup_directory != '\0')
-        gftp_set_directory (window1.request, startup_directory);
-
+      gftp_setup_startup_directory (window1.request);
       gftp_connect (window1.request);
       ftp_list_files (&window1, 0);
     }
@@ -1192,7 +1280,20 @@ main (int argc, char **argv)
   GDK_THREADS_ENTER ();
   gtk_main ();
   GDK_THREADS_LEAVE ();
+
   return (0);
 }
 
 
+void
+gftpui_show_or_hide_command (void)
+{
+  intptr_t cmd_in_gui;
+
+  gftp_lookup_global_option ("cmd_in_gui", &cmd_in_gui);
+  if (cmd_in_gui)
+    gtk_widget_show (gftpui_command_toolbar);
+  else
+    gtk_widget_hide (gftpui_command_toolbar);
+}
+
