@@ -18,122 +18,7 @@
 /*****************************************************************************/
 
 #include <gftp-gtk.h>
-static const char cvsid[] = "$Id: menu-items.c,v 1.24 2003/08/12 01:05:02 masneyb Exp $";
-
-void
-change_setting (gftp_window_data * wdata, int menuitem, GtkWidget * checkmenu)
-{
-  switch (menuitem)
-    {
-    case 1:
-      gftp_set_global_option ("ascii_transfers", GINT_TO_POINTER(1));
-      break;
-    case 2:
-      gftp_set_global_option ("ascii_transfers", GINT_TO_POINTER(0));
-      break;
-    case 3:
-      current_wdata = &window1;
-      other_wdata = &window2;
-      if (wdata->request)
-        update_window_info ();
-      break;
-    case 4:
-      current_wdata = &window2;
-      other_wdata = &window1;
-      if (wdata->request)
-        update_window_info ();
-      break;
-    }
-}
-
-
-void
-tb_openurl_dialog (gpointer data)
-{
-  const char *edttxt;
-
-  if (current_wdata->request->stopable)
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("%s: Please hit the stop button first to do anything else\n"),
-               _("OpenURL"));
-      return;
-    }
-
-  edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (hostedit)->entry));
-  if (*edttxt == '\0')
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("OpenURL: Operation canceled...you must enter a string\n"));
-      return;
-    }
-
-  if (GFTP_IS_CONNECTED (current_wdata->request))
-    disconnect (current_wdata);
-  else if (edttxt != NULL && *edttxt != '\0')
-    toolbar_hostedit (NULL, NULL);
-  else
-    openurl_dialog (current_wdata);
-}
-
-
-static void
-do_openurl (gftp_window_data * wdata, gftp_dialog_data * ddata)
-{
-  const char *tempstr;
-
-  if (current_wdata->request->stopable)
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("%s: Please hit the stop button first to do anything else\n"),
-               _("OpenURL"));
-      return;
-    }
-
-  tempstr = gtk_entry_get_text (GTK_ENTRY (ddata->edit));
-  if (*tempstr == '\0')
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("OpenURL: Operation canceled...you must enter a string\n"));
-      return;
-    }
-
-  if (GFTP_IS_CONNECTED (wdata->request))
-    disconnect (wdata);
-
-  if (gftp_parse_url (wdata->request, tempstr) == 0)
-    {
-      gtk_widget_destroy (ddata->dialog);
-      ftp_connect (wdata, wdata->request, 1);
-    }
-  else
-    gtk_widget_destroy (ddata->dialog);
-
-  ddata->dialog = NULL; 
-}
-
-
-void
-openurl_dialog (gpointer data)
-{
-  MakeEditDialog (_("Connect via URL"), _("Enter ftp url to connect to"),
-                  NULL, 1, NULL, gftp_dialog_button_connect, do_openurl, data,
-                  NULL, NULL);
-}
-
-
-void 
-disconnect (gpointer data)
-{
-  gftp_window_data * wdata;
-
-  wdata = data;
-  gftp_delete_cache_entry (wdata->request, NULL, 1);
-  gftp_disconnect (wdata->request);
-  remove_files_window (wdata);
-  update_window (wdata);
-}
-
+static const char cvsid[] = "$Id: menu-items.c,v 1.33 2004/03/21 13:27:02 masneyb Exp $";
 
 static void
 dochange_filespec (gftp_window_data * wdata, gftp_dialog_data * ddata)
@@ -223,11 +108,14 @@ dosave_directory_listing (GtkWidget * widget, gftp_save_dir_struct * str)
     }
 
   for (templist = str->wdata->files; 
-       templist->next != NULL;
+       templist != NULL;
        templist = templist->next)
     {
       tempfle = templist->data;
 
+      if (!tempfle->shown)
+        continue;
+
       tempstr = gftp_gen_ls_string (tempfle, NULL, NULL);
       fprintf (fd, "%s\n", tempstr);
       g_free (tempstr);
@@ -351,121 +239,6 @@ deselectall (gpointer data)
 }
 
 
-static void
-dosite (gftp_window_data * wdata, gftp_dialog_data * ddata)
-{
-  const char *edttext;
-
-  edttext = gtk_entry_get_text (GTK_ENTRY (ddata->edit));
-  if (*edttext == '\0')
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("SITE: Operation canceled...you must enter a string\n"));
-      return;
-    }
-
-  if (check_reconnect (wdata) < 0)
-    return;
-  gftp_site_cmd (wdata->request, edttext);
-
-  if (!GFTP_IS_CONNECTED (wdata->request))
-    disconnect (wdata);
-}
-
-
-void
-site_dialog (gpointer data)
-{
-  gftp_window_data * wdata;
-
-  wdata = data;
-  if (!check_status (_("Site"), wdata, 0, 0, 0, wdata->request->site != NULL))
-    return;
-
-  MakeEditDialog (_("Site"), _("Enter site-specific command"), NULL, 1,
-                  NULL, gftp_dialog_button_ok, dosite, wdata, NULL, NULL);
-}
-
-
-static void *
-do_change_dir_thread (void * data)
-{
-  int success, sj, network_timeout;
-  gftp_window_data * wdata;
-
-  wdata = data;
-
-  if (wdata->request->use_threads)
-    {
-      sj = sigsetjmp (jmp_environment, 1);
-      use_jmp_environment = 1;
-    }
-  else
-    sj = 0;
-
-  gftp_lookup_request_option (wdata->request, "network_timeout", 
-                              &network_timeout);
-
-  success = 0;
-  if (sj == 0) 
-    {
-      if (network_timeout > 0)
-        alarm (network_timeout);
-      success = gftp_set_directory (wdata->request, wdata->request->directory);
-      alarm (0);
-    }
-  else
-    {
-      gftp_disconnect (wdata->request);
-      wdata->request->logging_function (gftp_logging_error,
-                                        wdata->request,
-                                        _("Operation canceled\n"));
-    }
-
-  if (wdata->request->use_threads)
-    use_jmp_environment = 0;
-
-  wdata->request->stopable = 0;
-  return (GINT_TO_POINTER (success));
-}
-
-
-static int
-do_change_dir (gftp_window_data * wdata, char *directory)
-{
-  char *olddir;
-  int ret;
-
-  if (directory != wdata->request->directory)
-    {
-      olddir = wdata->request->directory;
-      wdata->request->directory = g_strdup (directory);
-    }
-  else
-    olddir = NULL;
-
-  ret = GPOINTER_TO_INT (generic_thread (do_change_dir_thread, wdata));
-
-  if (!GFTP_IS_CONNECTED (wdata->request))
-    {
-      disconnect (wdata);
-      if (olddir != NULL)
-        g_free (olddir);
-      return (-2);
-    }
-
-  if (ret != 0)
-    {
-      g_free (wdata->request->directory);
-      wdata->request->directory = olddir;
-    }
-  else
-    g_free (olddir);
-
-  return (ret);
-}
-
-
 int
 chdir_edit (GtkWidget * widget, gpointer data)
 {
@@ -474,81 +247,29 @@ chdir_edit (GtkWidget * widget, gpointer data)
   char *tempstr;
 
   wdata = data;
+  if (!check_status (_("Chdir"), wdata, gftpui_common_use_threads (wdata->request), 0, 0,
+                     wdata->request->chdir != NULL))
+    return (0);
+
+  if (check_reconnect (wdata) < 0)
+    return (0);
+
   edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (wdata->combo)->entry));
+
   if (!GFTP_IS_CONNECTED (wdata->request) && *edttxt != '\0')
     {
       toolbar_hostedit (NULL, NULL);
       return (0);
     }
 
-  if (!check_status (_("Chdir"), wdata, wdata->request->use_threads, 0, 0, 
-                     wdata->request->chdir != NULL))
-    return (FALSE);
-
   if ((tempstr = expand_path (edttxt)) == NULL)
     return (FALSE);
 
-  if (check_reconnect (wdata) < 0)
-    return (FALSE);
-
-  if (do_change_dir (wdata, tempstr) == 0)
-    {
-      gtk_clist_freeze (GTK_CLIST (wdata->listbox));
-      remove_files_window (wdata);
-      ftp_list_files (wdata, 1);
-      gtk_clist_thaw (GTK_CLIST (wdata->listbox));
-      add_history (wdata->combo, wdata->history, wdata->histlen, tempstr);
-    }
+  if (gftpui_run_chdir (wdata, tempstr))
+    add_history (wdata->combo, wdata->history, wdata->histlen, edttxt);
 
   g_free (tempstr);
-  return (FALSE);
-}
-
-
-int
-chdir_dialog (gpointer data)
-{
-  GList * templist, * filelist;
-  gftp_window_data * wdata;
-  char *newdir, *tempstr;
-  gftp_file *tempfle;
-  int num, ret;
-
-  wdata = data;
-  if (!check_status (_("Chdir"), wdata, wdata->request->use_threads, 1, 0, 
-                     wdata->request->chdir != NULL))
-    return (0);
-
-  filelist = wdata->files;
-  templist = GTK_CLIST (wdata->listbox)->selection;
-  num = 0;
-  templist = get_next_selection (templist, &filelist, &num);
-  tempfle = filelist->data;
-
-  newdir = gftp_build_path (wdata->request->directory, tempfle->file, NULL);
-
-  if ((tempstr = expand_path (newdir)) == NULL)
-    {
-      g_free (newdir);
-      return (0);
-    }
-
-  g_free (newdir);
-
-  if (check_reconnect (wdata) < 0)
-    return (0);
-
-  ret = 0; 
-  if (do_change_dir (wdata, tempstr) == 0)
-    {
-      gtk_clist_freeze (GTK_CLIST (wdata->listbox));
-      remove_files_window (wdata);
-      ftp_list_files (wdata, 1);
-      gtk_clist_thaw (GTK_CLIST (wdata->listbox));
-      ret = 1;
-    }
-  g_free (tempstr);
-  return (ret);
+  return (0);
 }
 
 
@@ -567,7 +288,7 @@ clearlog (gpointer data)
   textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (logwdw));
   len = gtk_text_buffer_get_char_count (textbuf);
   gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, len - 1);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, len);
   gtk_text_buffer_delete (textbuf, &iter, &iter2);
 #endif
 }
@@ -595,7 +316,6 @@ viewlog (gpointer data)
       return;
     }
   chmod (tempstr, S_IRUSR | S_IWUSR);
-  unlink (tempstr);
   
 #if GTK_MAJOR_VERSION == 1
   textlen = gtk_text_get_length (GTK_TEXT (logwdw));
@@ -604,11 +324,12 @@ viewlog (gpointer data)
   textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (logwdw));
   textlen = gtk_text_buffer_get_char_count (textbuf);
   gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, textlen - 1);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, textlen);
   txt = gtk_text_buffer_get_text (textbuf, &iter, &iter2, 0);
 #endif
   pos = txt;
-  do 
+
+  while (textlen > 0)
     {
       if ((len = write (fd, pos, textlen)) == -1)
         { 
@@ -619,11 +340,13 @@ viewlog (gpointer data)
         }
       textlen -= len;
       pos += len;
-    } while (textlen > 0);
+    }
 
+  fsync (fd);
   lseek (fd, 0, SEEK_SET);
-  view_file (tempstr, fd, 1, 0, 0, 1, NULL, NULL);
+  view_file (tempstr, fd, 1, 1, 0, 1, NULL, NULL);
   close (fd);
+
   g_free (tempstr);
   g_free (txt);
 }
@@ -659,7 +382,7 @@ dosavelog (GtkWidget * widget, GtkFileSelection * fs)
   textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (logwdw));
   textlen = gtk_text_buffer_get_char_count (textbuf);
   gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, textlen - 1);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, textlen);
   txt = gtk_text_buffer_get_text (textbuf, &iter, &iter2, 0);
 #endif
 
@@ -868,7 +591,7 @@ about_dialog (gpointer data)
 			       no_license_agreement, -1);
 #else
               textlen = gtk_text_buffer_get_char_count (textbuf);
-              gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen - 1);
+              gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen);
               gtk_text_buffer_insert (textbuf, &iter, no_license_agreement, -1);
 #endif
 	      gtk_widget_show (dialog);
@@ -884,7 +607,7 @@ about_dialog (gpointer data)
 		       no_license_agreement, -1);
 #else
       textlen = gtk_text_buffer_get_char_count (textbuf);
-      gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen - 1);
+      gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen);
       gtk_text_buffer_insert (textbuf, &iter, no_license_agreement, -1);
 #endif
       gtk_widget_show (dialog);
@@ -901,7 +624,7 @@ about_dialog (gpointer data)
       gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL, buf, -1);
 #else
       textlen = gtk_text_buffer_get_char_count (textbuf);
-      gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen - 1);
+      gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen);
       gtk_text_buffer_insert (textbuf, &iter, buf, -1);
 #endif
     }
