@@ -18,7 +18,7 @@
 /*****************************************************************************/
 
 #include <gftp-gtk.h>
-static const char cvsid[] = "$Id: options_dialog.c,v 1.18 2003/08/12 01:05:02 masneyb Exp $";
+static const char cvsid[] = "$Id: options_dialog.c,v 1.21 2003/12/28 16:02:04 masneyb Exp $";
 
 static GtkWidget * proxy_list, * new_proxy_domain, * network1,
                  * network2, * network3, * network4, * netmask1, * netmask2, 
@@ -71,7 +71,7 @@ _gen_input_widget (gftp_options_dialog_data * option_data, char *label, char *ti
   if (tiptxt != NULL)
     {
       tooltip = gtk_tooltips_new ();
-      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), tempwid, tiptxt, NULL);
+      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), tempwid, _(tiptxt), NULL);
     }
 
   return (tempwid);
@@ -192,7 +192,7 @@ _print_option_type_textcombo (gftp_config_vars * cv, void *user_data, void *valu
   if (cv->comment != NULL)
     {
       tooltip = gtk_tooltips_new ();
-      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), combo, cv->comment, NULL);
+      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), combo, _(cv->comment), NULL);
     }
 
   return (combo);
@@ -256,11 +256,11 @@ _textcomboedt_toggle (GtkList * list, GtkWidget * child, gpointer data)
   textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widdata->text));
   len = gtk_text_buffer_get_char_count (textbuf);
   gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, len - 1);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, len);
   gtk_text_buffer_delete (textbuf, &iter, &iter2);
 
   len = gtk_text_buffer_get_char_count (textbuf);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter, len - 1);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter, len);
   gtk_text_buffer_insert (textbuf, &iter, insert_text, -1);
 #endif
 }
@@ -416,10 +416,11 @@ _print_option_type_textcomboedt (gftp_config_vars * cv, void *user_data, void *v
   if (cv->comment != NULL)
     {
       tooltip = gtk_tooltips_new ();
-      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), combo, cv->comment, NULL);
+      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), combo, _(cv->comment), NULL);
 
       tooltip = gtk_tooltips_new ();
-      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), textwid, cv->comment, NULL);
+      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), textwid, _(cv->comment), 
+                            NULL);
     }
 
   return (widdata);
@@ -461,7 +462,7 @@ _save_option_type_textcomboedt (gftp_config_vars * cv, void *user_data)
   textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widdata->text));
   len = gtk_text_buffer_get_char_count (textbuf);
   gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, len - 1);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, len);
   newstr = gtk_text_buffer_get_text (textbuf, &iter, &iter2, 0);
   freeit = 1;
 #endif
@@ -564,7 +565,8 @@ _print_option_type_checkbox (gftp_config_vars * cv, void *user_data, void *value
   if (cv->comment != NULL)
     {
       tooltip = gtk_tooltips_new ();
-      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), tempwid, cv->comment, NULL);
+      gtk_tooltips_set_tip (GTK_TOOLTIPS(tooltip), tempwid, _(cv->comment),
+                            NULL);
     }
 
   return (tempwid);
@@ -717,6 +719,7 @@ apply_changes (GtkWidget * widget, gpointer data)
 
   proxy_hosts->list = new_proxy_hosts;
   new_proxy_hosts = NULL;
+  gftpui_show_or_hide_command ();
 }
 
 
